/*
Copyright 2003/2004, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.
Contributors: Hartmut Ritter

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2003/2004/2005
*/

#include <io.h>
#include <stdlib.h>

#include "ScatterWeb.Diffusion.h"
#include "ScatterWeb.DiffusionParams.h"
#include "ScatterWeb.DiffusionPtypes.h"
#include "ScatterWeb.DiffusionRouting.h"
#include "ScatterWeb.Event.h"
#include "ScatterWeb.Management.h"

//externally defined variables and structs
extern struct GradientStruct GradientTable[GradientTableSize];
extern appconfig_t AppConfig;
extern int debug_level;

unsigned int interestVector=0;

//forward declarations
int dumpDataMessage(char, unsigned long int, unsigned int);
int checkDataInterestMatchSigned (int, unsigned long int);

//
int getTemperatureNow(unsigned long int* result)
{
  temp_t temperature;
  System_startConversion();
  //System_clStartConversion(); 
  temperature = System_getTemperature();
  *result = temperature.u;
  return 1;
}


//////////////
//sendInitialDataMessage:
//after receiving the first InterestMessage an initial data message is sent
//in order to accelerate the path formation
//////////////
void sendInitialDataMessage(index)
{
  unsigned long int result = 0;  
  time_t currentTime;
  switch (GradientTable[index].attribute)
  {
      case INTEREST_VIB:
        //result = sensorData.vibCount;
        break; 
     case INTEREST_TEMP:
        getTemperatureNow(&result);
        break;
      case INTEREST_LIGHT:
        break;
      case INTEREST_PIR:
        break; 
      default:
        if (debug_level > DEBUG_HIGH)
          Comm_print("[DBG4]unknown attribute in sendInitialDataMessage\r\n");
        //return 0;
        return;
        break;    
  }
  Time_getSystemTime(&currentTime);
  GradientTable[index].lastSendTime = currentTime;
  if (debug_level > DEBUG_HIGH)
    Comm_print("[DBG4] sending inital Data Message\r\n");
  sendDataMessage( GradientTable[index].src, GradientTable[index].attribute,  
                    result, DATA_MESSAGE_FLAG_INITIAL, Configuration.id);
  return;
}

/////////////
//handleInterestHeartbeat
//get a specific sensor value at a regular time interval regardless of its value
//but: if the minimum distance between to value transmission remains under the parameter interval, 
//transmission is delayed
/////////////
void handleInterestHeartbeat (index)
{
  unsigned long int result=0;
  unsigned long int delta = 0;
  time_t currentTime;
 
  if (!GradientTable[index].valid)
    return;
    
  if (!GradientTable[index].heartbeat)
    return; // the heartbeat was obviously deleted in the meantime
    
  switch (GradientTable[index].attribute){
      case INTEREST_VIB:
        result = 144;//sensorData.vibCount;
        break; 
     case INTEREST_TEMP:
        getTemperatureNow(&result);
        break;
      case INTEREST_LIGHT:
        break;
      case INTEREST_PIR:
        break; 
      default:
        break;    
  }
  
  Time_getSystemTime(&currentTime);
  delta = currentTime.secs - GradientTable[index].lastSendTime.secs;
  if (delta < (unsigned long int) GradientTable[index].interval)
  {      
    //delay packet transmission because otherwise the minimum distance between to packets would be violated
    Timers_add(GradientTable[index].heartbeat*1000-delta,handleInterestHeartbeat, index);
  }
  else
  {
    //store the current time in the respective entry of the gradient table
    GradientTable[index].lastSendTime = currentTime;
    sendDataMessage( GradientTable[index].src, GradientTable[index].attribute,  
                    result, GradientTable[index].flags, Configuration.id);              
    //re-initialize the heartbeat timer for the next transmission
    Timers_add(GradientTable[index].heartbeat*1000, handleInterestHeartbeat, index);
  }  
}

/////////////
//handleInterestInterval
//pull the values of those sensors that are not rising event-based
//for example there is no "oh, there is suddenly a temperature"-event,
//but instead temperature value must be pulled from the respective sensor
//this is done regularly by this function for a specific index of the GradientTable
/////////////
void handleInterestInterval (index)
{
  unsigned long int result=0;
  int match;
  
  if (!GradientTable[index].valid)
    return;// Interest was deleted in the meantime
  if (!GradientTable[index].interval)
    return; // an Interest Interval of zero is not supported

  if (GradientTable[index].attribute == INTEREST_TEMP)
  {
    getTemperatureNow(&result);
    match=checkDataInterestMatchSigned(index, result);
    if (match)
    { 
      Time_getSystemTime(&(GradientTable[index].lastSendTime));
      sendDataMessage(GradientTable[index].src, GradientTable[index].attribute,  
                      result, GradientTable[index].flags, Configuration.id);                    
    }
    Timers_add((GradientTable[index].interval*1000), handleInterestInterval, index);
  }
}

/////////////
//checkDataInterestMatchSigned
//compare the value of result with the value stored in the GradientTableEntry
//return 1 if the result matches the logical condition stored in the Entry,
//return 0 if not
// NOTE: special function to compare temperature values, assumption: 0x8000 indicates a sign
/////////////
int checkDataInterestMatchSigned (int GradientTableRow, unsigned long int result)
{ 
  int sign_result=1, sign_value=1;
  
  if (result & 0x8000)
  {
    sign_result= -1;
  }
  if (GradientTable[GradientTableRow].value & 0x8000)
    sign_value = -1;  

  switch (GradientTable[GradientTableRow].operation)
  {
    case INTEREST_OPERATION_GREATER_THAN:
      if (sign_result > sign_value)//result > 0, value <0
      {
        return 1;
      }
      if (sign_value > sign_result)//value > 0, result <0
      {                              
        return 0;
      }
      if (sign_result==1)//result and value > 0
      {
        if (((result&0x7F80)>>7) > (GradientTable[GradientTableRow].value))
        {
          return 1;
        }
      }
      else //result < 0 and value < 0
      {
        if (((result&0x7F80)>>7) < ((GradientTable[GradientTableRow].value&0x7F80)>>7))
        {
          return 1;
        }
      }          
      break;
    case INTEREST_OPERATION_SMALLER_THAN:
      if (sign_result > sign_value)//result > 0, value <0
      {                              
        return 0;
      }
      if (sign_value > sign_result)//value > 0, result <0
      {
        return 1;
      }
      if (sign_result==1)//result and value > 0
      {
        if (((result&0x7F80)>>7) < (GradientTable[GradientTableRow].value))
        {
          return 1;
        }
      }
      else
      {
        if (((result&0x7F80)>>7) > ((GradientTable[GradientTableRow].value&0x7F80)>>7))
        {
          return 1;
        }
      }          
      break;
    case INTEREST_OPERATION_GREATER_EQUAL:
      if (sign_result > sign_value)//result > 0, value <0
        return 1;
      if (sign_value > sign_result)//value > 0, result <0
        return 0;
      if (sign_result==1)//result and value > 0
      {
        if (((result&0x7F80)>>7)>= (GradientTable[GradientTableRow].value))
          return 1;
        else
          return 0;
      }          
      break;
    case INTEREST_OPERATION_SMALLER_EQUAL:
      if (sign_value > sign_result)//value > 0, result <0
        return 1;
      if (sign_result==1)//result and value > 0
      {
        if (((result&0x7F80)>>7) <= (GradientTable[GradientTableRow].value))
          return 1;
        else
          return 0;
      }
      break;
    case INTEREST_OPERATION_DONT_CARE:
        return 1;
        break;
    default:
      if (debug_level > DEBUG_HIGH)
        Comm_print("[DBG4]unknown operation\r\n");
      return 0;
  }
  return 0;
}

/////////////
//checkDataInterestMatch
//compare the value of result with the value stored in the GradientTableEntry
//return 1 if the result matches the logical condition stored in the Entry,
//return 0 if not
// NOTE: handles only unsigned values, for signed temperature values checkDataInterestMatchSigned is called
/////////////
int checkDataInterestMatch(int GradientTableRow, unsigned long int result, char attribute)
{ 
  if (attribute == INTEREST_TEMP)
    return checkDataInterestMatchSigned (GradientTableRow, result);

  switch (GradientTable[GradientTableRow].operation)
  {
    case INTEREST_OPERATION_GREATER_THAN:
      if (result > GradientTable[GradientTableRow].value)
        return 1;
      break;
    case INTEREST_OPERATION_SMALLER_THAN:
      if (result < GradientTable[GradientTableRow].value)
        return 1;
      break;      
    case INTEREST_OPERATION_GREATER_EQUAL:
      if (result >= GradientTable[GradientTableRow].value)
        return 1;
      break;
    case INTEREST_OPERATION_SMALLER_EQUAL:
      if (result <= GradientTable[GradientTableRow].value)
        return 1;
      break;
    case INTEREST_OPERATION_DONT_CARE:
      return 1;
      break;
    default:
      if (debug_level > DEBUG_HIGH)
        Comm_print("unknown operation\r\n");
      return 0;
  }
  return 0;
}

/////////////
//ScatterWeb.DiffusionDataHandling_answerInterest
//called if either a local sensorEvent fits to an Interest 
// or if an incoming data message was received
/////////////
int DiffusionDataHandling_answerInterest(char attribute, unsigned long int result, unsigned int orig, char data_flags) //the attribute fits to a local sensor with activity
{
  int found=0, i=0;
  unsigned char transmissionPower = Configuration.transmitPower;
  unsigned char transmissionBoost=transmissionPower;
  time_t currentTime;
  
  if (transmissionPower < 89)
    transmissionBoost+=10;
  else
    transmissionBoost = 99;
  
  //look who was interested
  for (i=0;i<GradientTableSize;i++)
  {
    if (GradientTable[i].valid==0) 
      continue; //skip not valid entries    
    if (GradientTable[i].attribute != attribute)
      continue; //skip if attribute does not fit
    if (    (orig == Configuration.id)
        && (!checkDataInterestMatch(i, result, attribute)) )
        //the dataMessage originated here but is not in an interesting range
    {
      if (debug_level > DEBUG_HIGH)
        Comm_print("[DBG4]locally generated data message, but not in range\r\n");
      if (!(data_flags &DATA_MESSAGE_FLAG_INITIAL))
        continue;
    }
    if (GradientTable[i].flags & INTEREST_FLAG_LOCAL_INTEREST)
      {
        if (1)//(!(data_flags & DATA_MESSAGE_FLAG_INITIAL))//ignore the initial data message
        {
          if (GradientTable[i].src == Configuration.id)
          {
            Data_greenOn();
            if (debug_level > DEBUG_HIGH)
              Comm_print("[DBG4]DataMessage fits to Local Interest\r\n"); 
            dumpDataMessage(attribute, result, orig);
            Data_greenOff();
            found = 1;
          }
          else 
          {
            if (debug_level > DEBUG_HIGH){
              Comm_print("[DBG4]DataMessage fits to stored gradient, sending to requestor "); 
              Comm_print("%i",((unsigned int) GradientTable[i].src));
              Comm_print("\r\n");
              }            
            //setFunkPower(transmissionBoost);
            Net_setTxPower(transmissionBoost);
            sendDataMessage( (unsigned int) GradientTable[i].src, attribute, 
                            (unsigned long int) result, data_flags, (unsigned int) orig);
            //setFunkPower(transmissionPower);              
            Net_setTxPower(transmissionPower);
            found=1;        
          }
        }
      }        
    else
    { 
      if (orig == Configuration.id)//the Data Message was locally generated, not received from a neighbour
      { 
        Time_getSystemTime(&currentTime);                 
        if ((currentTime.secs - GradientTable[i].lastSendTime.secs)< 200*(unsigned int)GradientTable[i].interval)//send DataMessage only if min interval is not violated
        {
          if (debug_level > DEBUG_HIGH)
            Comm_print("[DBG4]not sending locally generated data message due to interval limit\r\n");
          found =1;
          return found;
        }                                                                   
      }
      //setFunkPower(transmissionBoost);
      Net_setTxPower(transmissionBoost);
      GradientTable[i].lastSendTime = CURRENT_TIME;        
      sendDataMessage( (unsigned int) GradientTable[i].src, attribute,  
                        (unsigned long int) result, data_flags, (unsigned int) orig);
      Net_setTxPower(transmissionPower);                        
      found=1;
    }
  //for now it is assumed that there is only one sink
  //this differs from directed diffusion which allows different sinks at different network parts
  if (found)
    break;
  }
  return found;
}      

////////////
//DiffusionDataHandling_addSensorInterest
//called by InsertGradientAt in ScatterWeb.Diffusion.c
//manages the local response to a InterestMessage
//either activates the event reporting (for vibration, e.g.) or activates regular response (for temp., e.g.)
//returns 1 if successful, 0 if not
////////////
int DiffusionDataHandling_addSensorInterest(char flags, char attribute, unsigned int heartbeat, int entry)
  {
  //remember the requested sensor type in the interestVector and the announce flags
  if (!(flags & INTEREST_FLAG_LOCAL_INTEREST))//not a locally inserted interest 
  {
    interestVector =  interestVector | attribute;  
    AppConfig.announceFlags |= AF_DIFFUSION;
  
    switch (attribute)
    {
      case INTEREST_VIB:
        Data_vibrationOn(5000);//activate vibration sensor event
        break;
      case INTEREST_TEMP:
        handleInterestInterval(entry); 
        break;
      default:
        return 0;
        break;
    }      
  }  
  //start a timer in case that a regular response was requested ("heartbeat")      
  if ((heartbeat)&&(!(GradientTable[entry].flags & INTEREST_FLAG_LOCAL_INTEREST)))
    Timers_add(heartbeat*1000, handleInterestHeartbeat, entry);
  return 1;
  }

////////////
//DiffusionDataHandling_handleSensorEvents 
//is called by handleEvent in ScatterWeb.Event.c if an event occurred
////////////
int DiffusionDataHandling_handleSensorEvents(sdata_t* data)
{
  int attribute=0;
  int retval = 1;
  
  if (data->sensor == SENSOR_VIBRATION)
  {
    Comm_print ("vibration detected ");
    Comm_print(" value %i",data->value);  
  }
  
  switch (data->sensor)
  {
    case SENSOR_VIBRATION:
      attribute = INTEREST_VIB;
      break;
    //case SENSOR_LIGHT:
    //atribute = INTEREST_LIGHT;
    //break;
    case SENSOR_MOVEMENT:
      attribute = INTEREST_PIR;
      break;
    case SENSOR_TEMPERATURE:
      attribute = INTEREST_TEMP;
      break;
    case SENSOR_BUTTON:
    case SENSOR_MICROPHONE:
    case SENSOR_RC5:
    case SENSOR_VOLTAGE:
      if (debug_level > MEDIUM)
        Comm_print("sensor type currently not supported by ScatterWeb.Diffusion\r\n");
      retval = 0;
      break;
    default:
      if (debug_level > MEDIUM)
        Comm_print("unknown sensor type in DiffusionDataHandling_handleSensorEvents\r\n");
      retval = 0;
  }
  if (retval)
    retval = DiffusionDataHandling_answerInterest(attribute, (unsigned long int) data->value, Configuration.id, 0);
  return retval;
}

////////////
//dumpDataMessage dumps incoming Data Message
////////////
int dumpDataMessage(char attribute, unsigned long int result, unsigned int source)
{
  System_stopWatchdog();
  switch (attribute)
  {
    case INTEREST_VIB:
      Comm_print("[VIB: "); 
      Comm_print("%l",result);
      Comm_print("]");
      break;
    case INTEREST_LIGHT:
      Comm_print("[LUM: "); 
      Comm_print("%l",result);
      Comm_print("]");
      break;
    case INTEREST_PIR:
      Comm_print("[PIR: "); 
      Comm_print("%l", result);
      Comm_print("]");
      break; 
    case INTEREST_TEMP:
      Comm_print("[TMP: "); 
      if (result & 0x8000)
      {
        Comm_print("-");
        result = (result ^ 0xFFFF) +1;
      }
      Comm_print("%l",(result&0x7F80)>>7);
      Comm_print("]");
      break; 
    default:
      if (debug_level > DEBUG_MEDIUM)
        Comm_print("[DBG4]unknown DataMessage attribute\r\n");
      System_startWatchdog();
      return 0;        
  }        
  Comm_print("[SRC: ");
  Comm_print("%l",source); 
  Comm_print("]");
  Comm_print("\r\n"); 
  System_startWatchdog();
  return 1;
} 
