/*
Copyright 2003/2004, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.
Contributors: Hartmut Ritter

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2003/2004/2005
*/

#include <io.h>
#include <stdlib.h>
#include "ScatterWeb.Event.h"
#include "../../../../System/src/ScatterWeb.System.h"

#include "ScatterWeb.Diffusion.h"
#include "ScatterWeb.DiffusionParams.h"
#include "ScatterWeb.DiffusionPtypes.h"
#include "ScatterWeb.Event.h"
#include "ScatterWeb.DiffusionRouting.h"
#include "ScatterWeb.Management.h"

//variables
#define MAX_SIZE_ROUTING_TABLE 20
static struct RoutingEntryStruct RoutingEntry [MAX_SIZE_ROUTING_TABLE];

//externally defined stuff
extern int debug_level;
extern struct GradientStruct GradientTable[GradientTableSize];

//forward declarations
int do_sendRoutablePacket(unsigned int, unsigned int, unsigned int, unsigned char, unsigned char*, unsigned int);
int dumpRoutablePacketHeader(unsigned int*, unsigned int*, struct RoutablePacketHeaderStruct*);
int DiffusionRouting_lookupRoutingTable(unsigned int srcid, unsigned int final_dest, unsigned int* next_hop);

//////////////
//DiffusionRouting_init() is used for initialization, called by Diffusion_init()
//////////////
int DiffusionRouting_init()
{
  int i=0;  
  for (i=0;i<MAX_SIZE_ROUTING_TABLE;i++)
  {
    RoutingEntry[i].id = 0; 
    RoutingEntry[i].next_hop = 0;
    RoutingEntry[i].flags = 0;
  }
 return 1; 
}

//////////////
//nextHopFromGradientTable
//helper function used to find out if the requested id is known as a gradient source in the gradient table
//params: unsigned int* id  id of the node that is looked for
//returns 1 if id found as gradient source, 0 otherwise
//////////////
int nextHopFromGradientTable(unsigned int* id)
{
  int i = 0;
  for (i=0;i<GradientTableSize;i++)  //look for an already existing identical entry in Gradient Table
  {
    if (GradientTable[i].valid==0)
      continue;
    if (GradientTable[i].src == Configuration.id)//paranoid, but you never know...
      continue;    
    *id = GradientTable[i].src; //the first one, this is only o.k. if we have only one sink
    return 1;
  }
  return 0;
}

/*
int sendRoutingInfo(unsigned int id, unsigned int dest)
{  
  struct RouteInfoStruct RouteInfoPacket;

  RouteInfoPacket.id = id;
  RouteInfoPacket.flags = 0;
  Net_sendPacket(dest, ROUTING_MESSAGE, NULL, 0, (char*) &RouteInfoPacket, sizeof(RouteInfoPacket), NULL);

  if (debug_level > DEBUG_HIGH)              
    Comm_print("[DBG4]Route Info sent \r\n");   
  return 1;
}
*/

//////////
//sendRoutable sends a packet of tpye ROUTABLE_PACKET
//returns: 
//  1 if routing table entry was found and packet was sent, 
//  0 otherwise
/////////
int DiffusionRouting_sendRoutablePacket(packet_t* packet, fp_char_t callback)
{
return DiffusionRouting_sendRoutable(packet->from, packet->to, packet->type, packet->header, packet->header_length, packet->data, packet->data_length, callback);
}
//////////
//sendRoutable sends a packet of tpye ROUTABLE_PACKET
//returns: 
//  1 if routing table entry was found and packet was sent, 
//  0 otherwise
/////////
//orig_src, final_dest, inner_type, header, header_length, data, data_length, callback
int DiffusionRouting_sendRoutable(unsigned int orig_src, unsigned int final_dest, unsigned char inner_type, unsigned char* header, unsigned int header_length, unsigned char* data,unsigned int data_length, fp_char_t callback)
{
  unsigned int next_hop;
  
  if (DiffusionRouting_lookupRoutingTable(orig_src, final_dest, &next_hop))
  {
    do_sendRoutablePacket(next_hop, orig_src, final_dest, inner_type, data, data_length);
    return 1;
  }    
  else
    return 0;
}

//do_sendRoutablePacket really sends a routable packet
int do_sendRoutablePacket(unsigned int next_hop, unsigned int orig, unsigned int final_dest, unsigned char inner_type, unsigned char* data, unsigned int data_length)
{
  struct RoutablePacketHeaderStruct RoutablePacketHeader;

  RoutablePacketHeader.final_dest = final_dest;
  RoutablePacketHeader.orig = orig;
  RoutablePacketHeader.type = inner_type; 

  //Net_sendPacket(next_hop, ROUTABLE_PACKET, (unsigned char*) &RoutablePacketHeader, 
  //  sizeof(RoutablePacketHeader), (unsigned char*) data, data_length, NULL);  
  packet_t packet;
  packet.to = next_hop;
  packet.from = Configuration.id;
  packet.type = DD_ROUTABLE_PACKET;
  packet.header = (char*) &RoutablePacketHeader;
  packet.header_length = sizeof (RoutablePacketHeader);
  packet.data = (unsigned char*) data;
  packet.data_length = data_length;
  Net_send(&packet, NULL);

  return 1;
}

//////////
//updateRoutingTable is called each time a Data Message arrives at the node
//so the node learns from the Data Messages flowing downstream from the sources to the sink
//////////
int DiffusionRouting_updateRoutingTable(unsigned int id, unsigned int next_hop)
{
  int i=0;
  
  for (i=0;i<MAX_SIZE_ROUTING_TABLE;i++)
  {
    if (RoutingEntry[i].id == id)
    {
      RoutingEntry[i].next_hop = next_hop;
      return 1;       
    }                   
  }
  //at this point the id was not in the Routing Table
  //therefore look for an empty row and insert new entry
  for (i=0;i<MAX_SIZE_ROUTING_TABLE;i++)
  {
    if (RoutingEntry[i].id == 0)
    {
      RoutingEntry[i].id = id;
      RoutingEntry[i].next_hop = next_hop;
      return 1;       
    }                   
  }
  return 0;
}

//////////
//lookupRoutingTable performs a routing table lookup for a given destination id
//routing table entries are learned by peeking into downstream data messages
//returns: 
//  1 if routing table entry was found, 
//  0 otherwise
/////////
int DiffusionRouting_lookupRoutingTable(unsigned int srcid, unsigned int final_dest, unsigned int* next_hop)
//srcid: id of the last hop the packet came from
//final_dest: the final destination of the packet
//next_hop: if successful, this variable contains the next hop the packet should be sent to
{
  int i=0;
  unsigned int gradient_next_hop;
  
  for (i=0;i<MAX_SIZE_ROUTING_TABLE;i++)
  {
    if (final_dest == RoutingEntry[i].id)
    {
      *next_hop = RoutingEntry[i].next_hop;
      if (*next_hop == srcid)
        return 0;//paranoid, just to make definitely sure no ping-pong is played
      return 1;       
    }                   
  }
  //no routing entry in Routing Table found 
  //in that case the destination should be downstream
  //this is problematic if:
  //  - the node missed Data Messages, e.g. due to a reboot of the node
  //  - there is more than one sink

  if (!nextHopFromGradientTable(&gradient_next_hop))
    return 0;
  *next_hop = gradient_next_hop; 

  if (*next_hop == srcid)
    return 0;//paranoid, just to make definitely sure no ping-pong is played    

  //another check: make sure next_hop is not in the routing table
  //in that case something is inconsistent: a node can not be upstream AND downstream
  for (i=0;i<MAX_SIZE_ROUTING_TABLE;i++)
  {
    if (*next_hop == RoutingEntry[i].id)
      return 0;
  }
  return 1;
}

/////////////
//dumpRoutingTable() implements the command term_rrt: read routing table
/////////////
int dumpRoutingTable (unsigned int requestor)
{
  int i = 0;
  int found = 0;

  //first case: the command was issued on behalf of another node
  if (requestor != Configuration.id)//in that case send answer back as routable packets
  {
    System_stopWatchdog();
    struct RoutingEntryStruct RoutingEntryPacket;
    for (i=0;i<MAX_SIZE_ROUTING_TABLE;i++)
    {
      if (!RoutingEntry[i].id) continue; //skip entries with empty id's
      
      RoutingEntryPacket.id = RoutingEntry[i].id;
      found = 1;
      RoutingEntryPacket.next_hop = RoutingEntry[i].next_hop;
      RoutingEntryPacket.flags = RoutingEntry[i].flags | ROUTING_MESSAGE_FLAG_INFORMATIONAL;
      
      if (!DiffusionRouting_sendRoutable(Configuration.id, requestor, DD_ROUTING_MESSAGE, NULL, 0, (unsigned char*) &RoutingEntryPacket, (unsigned int) sizeof(RoutingEntryPacket),NULL))
      {
        System_startWatchdog();
        return 0;    
      }
    }
    if (!found)     
    {
      RoutingEntryPacket.id = 0;
      RoutingEntryPacket.next_hop = 0;
      RoutingEntryPacket.flags = ROUTING_MESSAGE_FLAG_INFORMATIONAL | ROUTING_MESSAGE_FLAG_NO_ENTRY;
      
      if (!DiffusionRouting_sendRoutable(Configuration.id, requestor, DD_ROUTING_MESSAGE, NULL, 0, (unsigned char*) &RoutingEntryPacket, (unsigned int) sizeof(RoutingEntryPacket), NULL))
      {
        System_startWatchdog();
        return 0;    
      }    
    }      
    System_startWatchdog();
    return 1;
  }  
 
  //second case: the command was issued locally
  int entry = 0;
  for (i=0;i<MAX_SIZE_ROUTING_TABLE;i++)
  {
    if (RoutingEntry[i].id)
    {
      entry = 1;    
      String_write(term_reply,"dest %i via next hop %i\r\n", RoutingEntry[i].id, RoutingEntry[i].next_hop);
    /*
      Comm_print("dest:");
      Comm_print("%i",RoutingEntry[i].id);
      Comm_print(" via next hop:");
      Comm_print("%i",RoutingEntry[i].next_hop);
      Comm_print("\r\n");
      */
    } 
    if (!entry)
      String_write(term_reply,"routing table empty");
  }
  return 1;
}

//////////
//handleRoutablePacket is the packet handling function for incoming packets of type ROUTABLE_PACKET
//////////
int DiffusionRouting_handleRoutablePacket(unsigned int* dstid, unsigned int* srcid, unsigned char* payload, unsigned int data_length) 
{
  //int i =0;
  //first case: the packet is not for us, only for debugging purposes
  if (*dstid != Configuration.id)
  {
    if (debug_level > DEBUG_HIGH)
      Comm_print("[DBG4]sniffed a routable packet not for me\r\n");
    return 0;  
  }
  struct RoutablePacketHeaderStruct* RoutablePacket;
  RoutablePacket = (struct RoutablePacketHeaderStruct*) payload;
  if (debug_level > DEBUG_HIGH)
  { 
    Comm_print("[DBG4]"); 
    dumpRoutablePacketHeader(srcid, dstid, RoutablePacket);
    Comm_print("\r\n");
  }
  //second case: the packet was sent to this node, but should be forwarded
  if (RoutablePacket->final_dest != Configuration.id) //try to forward the packet
  {    
    unsigned int next_hop; 
    
    if (debug_level > DEBUG_HIGH)
      Comm_print("[DBG4]trying to forward the packet..");
    
    if (DiffusionRouting_lookupRoutingTable(*srcid, RoutablePacket->final_dest, &next_hop))
    {
      if (debug_level > DEBUG_HIGH)
      {
        Comm_print(" using node ");
        Comm_print("%i",next_hop);
        Comm_print("\r\n");        
      }    
      
      packet_t p;
      p.to = next_hop;
      p.type = DD_ROUTABLE_PACKET;
      p.header = NULL;
      p.header_length = 0;
      p.data = (unsigned char*) payload;
      p.data_length = data_length;
      Net_send(&p, NULL);
        
      //sendPacket(next_hop, ROUTABLE_PACKET, NULL, 0, (unsigned char*)payload, data_length, NULL);
      //(unsigned int to, unsigned char type, unsigned char* header, unsigned int header_length, 
      // unsigned char* data, unsigned int data_length, fp_char_t callback)    
      return 1;
    }
    else
    {
      if (debug_level > DEBUG_HIGH)
      Comm_print("..forw. failed, no idea of next hop\r\n");
      return 0;
    }
  }  
  //third case: this node is endpoint of the packet, so handle it
  //struct TermCmdPacketStruct* TermCmdPacket;
  struct InterestMessageInfoStruct* InterestMessageInfoPacket;
  struct RoutingEntryStruct* RoutingEntryPacket;
  unsigned char* inner_payload;
  int passed = 0; 
  switch (RoutablePacket->type)
  {
    case MESSAGING_REQ_PACKET:
      inner_payload = (unsigned char*) &payload[sizeof(*RoutablePacket)];
      Diffusion_routableRequestHandler (RoutablePacket->orig, inner_payload);
      break;
    case MESSAGING_ANS_PACKET:
      inner_payload = (unsigned char*) &payload[sizeof(*RoutablePacket)];
      Diffusion_routableAnswerHandler(RoutablePacket->orig, inner_payload);
      break;
    case DD_INTEREST_MESSAGE:
      InterestMessageInfoPacket = (struct InterestMessageInfoStruct*) &payload[sizeof(*RoutablePacket)];
      
      if ((InterestMessageInfoPacket->flags & INTEREST_FLAG_INFORMATIONAL)
          || (debug_level > DEBUG_MEDIUM))
        //dumpInterestMessageInfoPacket(srcid, InterestMessageInfoPacket);
        dumpInterestMessageInfoPacket(&(RoutablePacket->orig), InterestMessageInfoPacket);
      
      if (InterestMessageInfoPacket->flags & INTEREST_FLAG_COMMAND)
      {       
        struct InterestMessageStruct InterestMessagePacket;
        copyInterestMessageInfoToInterestMessage(&InterestMessagePacket, InterestMessageInfoPacket);
        passed = consistencyCheck(InterestMessagePacket);
        if (passed)
        {
          if (debug_level > DEBUG_HIGH)
          {
            Comm_print("[DBG4]Modifying interest due to remote command");
          }
          changeInterestParams(InterestMessagePacket);         
        }      
      }        
      break;
    /*  
    case TERM_CMD:
      TermCmdPacket = (struct TermCmdPacketStruct*) &payload[sizeof(*RoutablePacket)];
      if (debug_level > DEBUG_HIGH)
      {
        Comm_print("[DBG4]rcvd. TERM CMD, ");
        Comm_print("subcmd ");
        Comm_print("%i",TermCmdPacket->subcmd);
        Comm_print("\r\n");
      }
      //copy the command arguments        
      for (i=0;i<10;i++)
      {
        command[4+i]=TermCmdPacket->command[i];
      }
      //execute the appropriate command from the term_commands list
      term_commands[TermCmdPacket->subcmd].function(RoutablePacket->orig);      
      break;    

*/
    case DD_ROUTING_MESSAGE:
      RoutingEntryPacket = (struct RoutingEntryStruct*) &payload[sizeof(*RoutablePacket)];
      if (RoutingEntryPacket->flags & ROUTING_MESSAGE_FLAG_INFORMATIONAL)      
      {
        Comm_print("[RES rrt][ID: ");
        Comm_print("%i",RoutablePacket->orig);
        Comm_print("]");
        if (RoutingEntryPacket->flags & ROUTING_MESSAGE_FLAG_NO_ENTRY)
          Comm_print("[N/A]\r\n");
        else
        {
          Comm_print("[DST: ");
          Comm_print("%i",RoutingEntryPacket->id); 
          Comm_print("][VIA: ");
          Comm_print("%i",RoutingEntryPacket->next_hop);
          Comm_print("]\r\n");
        }
      }  
      break;                          
    default:
      if (debug_level > DEBUG_HIGH)
        Comm_print("[DBG4]ERROR: received routable packet of unknown type\r\n");
      break;
  }
  return 1; 
}

/////////////////////////////////////////////
// TERMINAL COMMANDS
/////////////////////////////////////////////

///////////
//term_rtt is the terminal command to read the routing table
///////////
COMMAND(rrt,0)
{ 
    dumpRoutingTable(Configuration.id);
    return;
}

////////////////////////////////////////////////////////////////////////////////
//DEBUG FUNCTIONS
////////////////////////////////////////////////////////////////////////////////

///////////
//dumpRoutablePacketHeader dumps the header fields of a routable packet
///////////
int dumpRoutablePacketHeader(unsigned int* srcid, unsigned int* dstid, struct RoutablePacketHeaderStruct* RoutablePacket)
{
    Comm_print("packet from");
    Comm_print("%i",*srcid);
    Comm_print(" to");
    Comm_print("%i",*dstid);
    Comm_print(" orig src");
    Comm_print("%i",RoutablePacket->orig);
    Comm_print(" final dest");
    Comm_print("%i",RoutablePacket->final_dest);
    Comm_print(" type");
    Comm_print("%i",RoutablePacket->type);
    return 1;
}



