/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

/** \file ScatterWeb.RSA.h
 ** \ingroup Applications
 ** \brief Methods for asymmetric ciphering.
 ** <p>
 ** Use RSA:
 ** <ul> 
 ** <li>call RSA_init() to genetare your own keys
 ** </ul> 
 ** <p> 
 ** Use encoding:
 ** <ul> 
 ** <li>call RSA_getPublicKey(int id) to get the the public key of the other node
 ** <li>wait until RSA_isPublicKeyReceived() returns True
 ** <li>call RSA_encodePayload(packet_t* packet) to encode the payload of the packet
 ** </ul> 
 ** <p>
 ** Use decoding:
 ** <ul> 
 ** <li>call RSA_decodePayload(packet_t* packet) to decode the payload of the packet 
 ** </ul> 
 **
 ** @author        Marco Hutta
 ** @version       2.2
 ** @since         2.2
 **/ 

#ifndef ScatterWeb_RSA
#define ScatterWeb_RSA

#include "../../../../System/src/ScatterWeb.System.h"

/**
 * generates private and public keys
 *
 * @return        <code>True</code> always.
 * @since         2.2
 */
bool RSA_init(); 

/**
 * gets the public key
 *
 * @param id      id of the node
 * @return        <code>True</code> always.
 * @since         2.2
 */
bool RSA_getPublicKey(int id);

/**
 * checks the receiving of the public key
 *
 * @return        <code>True</code> if public key received;
 *                <code>False</code> otherwise.
 * @since         2.2
 */
bool RSA_isPublicKeyReceived();

/**
 * encode the payload field of the packet
 *
 * @param packet  packet with payload to encode
 * @return        pointer of packet
 * @since         2.2
 */
packet_t* RSA_encodePayload(packet_t* packet);

/**
 * decodes the payload field of the packet
 *
 * @param packet  packet with payload to decode
 * @return        pointer of packet
 * @since         2.2
 */
packet_t* RSA_decodePayload(packet_t* packet) ;

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool RSA_keyRequestHandler();

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool RSA_keyReplyHandler(); 

#endif
