/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

/** \file ScatterWeb.SMS.h
 ** \ingroup Applications
 ** \brief  This interface will support sending and receiving information via
 ** mobile phones for watching and controlling the ScatterWeb environment over
 ** large distances.  
 ** <p> 
 ** Features:
 ** <ul>
 ** <li>Text messaging
 ** <li>Setting MIN (if required)
 ** <li>Receiving text messages
 ** </ul>
 ** <p>
 ** Preparation:
 ** <ul>
 ** <li>Connect the ESB to a cellular phone (tested with Siemens S65)
 ** </ul>
 ** <p>
 ** Using: 
 ** Init - call SMS_init(char min[4]) by yourself:
 **                      -> setState(AT_WAIT)            -> send("AT")
 ** -> if answer OK      -> setState(AT_OK)              -> call isMINRequired()
 **                      -> setState(AT_CPIN_WAIT)       -> send("AT+CPIN")
 **    else ...
 ** -> if answer READY   -> setState(CONNECT)
 **    else SIM PIN      -> setState(AT_CPIN_XXXX_WAIT)  -> call setMin()
 **                      -> setState(AT_CPIN_XXXX_WAIT)  -> send("AT+CPIN=...")
 ** -> if answer OK      -> setState(CONNECT)
 **    else answer ERROR -> setState(AT_CPIN_XXXX_ERROR)
 **                                                      -> return True 
 ** <p>
 ** Get number of messages - call SMS_registerNumberCallback(fp_t fp) to
 ** register a function that except to pass the number of messages and call
 ** SMS_isMessageAvailable() by yourself:  
 ** if state != CONNECT                                  -> return False 
 **                      -> setState(AT_CPMS_WAIT)       -> send("AT+CPMS=...")
 ** -> if answer +CPMS   -> setState(CONNECT)            -> call numberCallback()
 **                                                      -> return True 
 ** <p>
 ** Get the message - call SMS_registerMessageCallback(fp_t fp) to
 ** register a function that except to pass the struct of a sms and call 
 ** SMS_getMessage(int number) by yourself:
 ** if state != CONNECT                                  -> return False  
 **                      -> setState(AT_CMGR_WAIT)       -> send("AT+CMGR=...")
 ** -> if answer +CMGR   -> setState(CONNECT)            -> call messageCallback()
 **                                                      -> return True 
 ** <p>
 ** Delete the message - call SMS_deleteMessage(int number) by yourself:
 ** if state != CONNECT                                  -> return False  
 **                      -> setState(AT_CMGD_WAIT)       -> send("AT+CMGD=...")
 ** -> if answer OK      -> setState(CONNECT)
 **                                                      -> return True 
 ** <p>
 ** Send message - call SMS_sendMessage(char* number, char* text) by yourself:
 ** if state != CONNECT                                  -> return False 
 **                                                      -> send("AT+CMGF=...")
 **                                                      -> send("AT+CMGS=...")
 **                                                      -> send(msg)
 **                                                      -> send((UINT8*)'\26')
 **                                                      -> return True
 ** <p>
 ** Check connection - call SMS_isConnected() by yourself:
 ** if state == AT_WAIT                                  -> return False
 ** if state == AT_OK                                    -> return False
 ** if state == AT_CPIN_WAIT                             -> return False
 ** if state == AT_CPIN_XXXX_WAIT                        -> return False
 ** if state == AT_CPIN_XXXX_ERROR                       -> return False
 **                                                      -> return True 
 **
 ** @author        Marco Hutta
 ** @version       2.2
 ** @since         2.0
 **/ 

#ifndef ScatterWeb_SMS
#define ScatterWeb_SMS

#include "../../../../System/src/ScatterWeb.System.h"

/**
 * Struct for the sms
 * <ul>
 * <li>number - number of sender
 * <li>message - text of sms
 * </ul>
 *  
 * @since         2.2
 */
typedef struct {
  UINT8* number;
  UINT8* message;
} sms_t;

/**
 * struct for received sms
 *
 * @since         2.2
 */
extern volatile sms_t rxSMS;

int SMS_numberOfSms;

/**
 * Initializes the sms module und sets the MIN if required.
 *
 * @param min     MIN else "0000" 
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool SMS_init(char min[4]);

/**
 * Checks state of connection.
 *
 * @return        <code>True</code> if mobile is connected;
 *                <code>False</code> otherwise.
 * @since         2.2
 */
bool SMS_isConnected();

/**
 * Checks number of messages.
 *
 * @return        <code>True</code> always.
 * @see           #SMS_registerNumberCallback(fp_t fp)
 * @since         2.2
 */
bool SMS_isMessageAvailable();

/**
 * Gets message.
 *
 * @parm number   position of message
 * @return        <code>True</code> if connected;
 *                <code>False</code> otherwise.
 * @see           #SMS_registerMessageCallback(fp_t fp)
 * @since         2.2
 */
bool SMS_getMessage(int number);

/**
 * Deletes message.
 *
 * @parm number   position of message
 * @return        <code>True</code> if connected;
 *                <code>False</code> otherwise.
 * @since         2.2
 */
bool SMS_deleteMessage(int number);

/**
 * Sends messages.
 *
 * @parm number   number of receiver
 * @param text    text of message 
 * @return        <code>True</code> if connected;
 *                <code>False</code> otherwise.
 * @since         2.0
 */
bool SMS_sendMessage(char* number, char* text);

/**
 * Registers to get messages
 *
 * @param fp      function to call
 * @return        <code>True</code> always.
 * @see           #SMS_getMessage(int number)
 * @since         2.2
 */
bool SMS_registerMessageCallback(fp_t fp);

/**
 * Registers to get number
 *
 * @param fp      function to call
 * @return        <code>True</code> always.
 * @see           #SMS_isMessageAvailable();
 * @since         2.2
 */
bool SMS_registerNumberCallback(void(*fp_t)());

#endif
