/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

/** \file ScatterWeb.Event.h
 ** \ingroup Applications
 ** \brief Methods for handling and forwarding sensor data events.
 ** <p>
 ** Processes events according to sensorMask and announceFlags:
 ** <p>
 ** appconfig_t.sensorMask contains what sensors will be processed (commands "rsm"/"ssm"):
 ** <ul>
 ** <li>0x10 SENSOR_VIBRATION - Vibration sensor
 ** <li>0x08 SENSOR_MOVEMENT - PIR sensor
 ** <li>0x80 SENSOR_BUTTON - Button
 ** <li>0x01 SENSOR_MICROPHONE - Microphone
 ** <li>0x02 SENSOR_TEMPERATURE - Temperature
 ** <li>0x04 SENSOR_RC5 - RC5 Receiving
 ** <li>0x40 SENSOR_VOLTAGE - Battery Voltage
 ** </ul>
 ** <p>
 ** appconfig_t.announceFlags contains what will be done with sensor events (commands "raf"/"saf"):  
 ** <ul>
 ** <li> 0x01 AF_SERIAL - Prints events to serial port.
 ** <li> 0x02 AF_RADIO - Broadcasts events in a ::SENSOR_PACKET.
 ** </ul>
 ** <p>
 ** So to get all sensors reported over serial link and radio do:
 ** \code
 ** ssm FF
 ** saf 03
 ** \endcode
 **
 ** @author        Marco Hutta
 ** @author        Bjoern Lichtblau
 ** @version       2.2
 ** @since         2.0
 **/ 
    
#ifndef ScatterWeb_Event
#define ScatterWeb_Event

#include "../../../../System/src/ScatterWeb.System.h"

/**
 * Prints events to serial port.
 *
 * @see           #appconfig_t.announceFlags
 * @since         2.0
 */
#define AF_SERIAL    (0x01)

/**
 * Broadcasts events in a SENSOR_PACKET.
 *
 * @see           #appconfig_t.announceFlags
 * @since         2.0
 */
#define AF_RADIO     (0x02)

/**
 * Application config type.
 * <ul>
 * <li>announceFlags - If events occur this variable says where they get announced.
 * <li>sensorMask - What local events will be handled by the ehandler.
 * </ul>
 *  
 * @since         2.0
 */
typedef struct {
  UINT8 announceFlags;
  UINT8 sensorMask;
} appconfig_t;

/**
 * Initializes the event handling
 *
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool Event_init();

#endif
