/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

#include "ScatterWeb.DHCP.h"

#ifdef DHCPSERVER
typedef struct _scope {
  UINT8 timeout; // timeout: 0=do revoke address
  UINT8 free:1;  // 1=address available; 0=not free
} scope;

scope addresspool[MAX_ADDRESSES]; // place where all addresses are saved
#endif

int dhcp_address_received; // was an address received from the server? 1=yes; 0=no
UINT32 current_xid;        // client saves its xid@_edHCP_Header; random session number
UINT32 mac_address;        // unique address for this node (randomly generated)
UINT16 server_address;     // fix address of server

typedef struct _eDHCP_Header{
  UINT8 op;      // 0=client, 1=server
  UINT8 hlen;    // length of hardware address
  UINT8 hops;    // number of hops between node A and node B
  UINT32 xid;    // random session number
  UINT16 flags;  // status bits
  UINT16 ciaddr; // address of client
  UINT16 yiaddr; // assigned address from server
  UINT16 siaddr; // server address
  UINT16 chaddr; // mac address
  char* sname;   // Name of server (optional)
  UINT8 msgtype; // user defined: DHCPDISCOVER,...
}eDHCP_Header;

// msgtypes
#define DHCPDISCOVER        0 // client to server : give me an address
#define DHCPOFFER           1 // server to client : here is your address
#define SERVER_DHCPREVOKE   2 // server to client : address expired (timeout)
#define CLIENT_DHCPREVOKE   3 // client to server : address refresh
#define SERVER_DHCPACOLLECT 4 // server to client : collection of all client adresses
#define CLIENT_DHCPACOLLECT 5 // client to server : send client address

// set standard parameter in DHCP_Header packet. Useful for reserved parameters, e.g. flags
void defaultHeader(eDHCP_Header *e) {
  e->hlen = 2;
  e->hops = 1;
  e->flags = 0;
  e->siaddr = 1;
  e->chaddr = mac_address;
  e->sname = 0;
}

// generates a pseudo random 32 Bit number
UINT32 __random32() {
  UINT32 res = 0;
  int i;
  for (i=0; i<4; ++i) {
    res = res << 8;
    res += ((ADC12MEM5+ADC12MEM4+ADC12MEM3+ADC12MEM2+ADC12MEM1+ADC12MEM0) & 0x00FF);
  }
  return res;
}

// sends an dhcp_packet to a specified address over radio
// Parameter:  eDHCP_Header - the struct to be send UINT16 - address to be send to
void sendDHCPPacket(eDHCP_Header *dhcp_packet, UINT16 toaddress) {
  packet_t packet;
  packet.to = toaddress;
  packet.type = DHCP_PACKET;
  packet.header = NULL;
  packet.header_length = 0;
  packet.data = (UINT8*) dhcp_packet;
  packet.data_length = sizeof(eDHCP_Header);
  Net_send(&packet, NULL);
}

#ifdef DHCPCLIENT
void wait_client() {
  // if address was not received from server then set ID to current_xid
  if (dhcp_address_received == 0) {
    Configuration.id = (UINT16)current_xid;
    Configuration_save();
    dhcp_address_received = 2;
  }
}
#endif

#ifdef DHCPSERVER
// Each node has a timeout saved server-side. If timeout "nearly"
// expires, then this function sends a message to the corresponding client
// and the client may send its address to the server, where timeout restarts.
void timeout_f() {
  int i;
  for (i=0; i<MAX_ADDRESSES; ++i) {
    if (addresspool[i].timeout==5 && addresspool[i].free==0) {
      // if timeout has "nearly" expired, then send DHCPREVOKE to client
      eDHCP_Header dhcp_packet;
      defaultHeader(&dhcp_packet);
      dhcp_packet.op = 1;
      dhcp_packet.xid = 0;
      dhcp_packet.ciaddr = i+2;
      dhcp_packet.yiaddr = 0;
      dhcp_packet.msgtype = SERVER_DHCPREVOKE;
      sendDHCPPacket(&dhcp_packet, dhcp_packet.ciaddr);
      Comm_print("DHCP_SERVER: Address %i nearly expired.\r\n",i+2);
    } else if (addresspool[i].timeout==0 && addresspool[i].free==0) {
      // if timeout has expired, then delete the client from internal server address pool
      addresspool[i].free = 1;
      Comm_print("DHCP_SERVER: Address %i expired.\r\n",i+2);
    }
    --addresspool[i].timeout;
  }
  Timers_add(TIMEOUT, timeout_f, 0);
}
#endif

void startDHCP() {
  eDHCP_Header dhcp_packet;
  #ifdef DHCPCLIENT
  Comm_print("DHCP_CLIENT: DHCPDISCOVER. Send ID request to server.\r\n");
  // try to get an address from dhcp-server.
  defaultHeader(&dhcp_packet);
  dhcp_packet.op = 0;
  dhcp_packet.xid = current_xid = __random32();
  dhcp_packet.ciaddr = Configuration.id; // current ID/Address
  dhcp_packet.yiaddr = 0;
  dhcp_packet.msgtype = DHCPDISCOVER;
  sendDHCPPacket(&dhcp_packet, BROADCAST);
  // wait some seconds for DHCP-Answer from Server
  Timers_add(8000, wait_client, 0);
  #endif
  #ifdef DHCPSERVER
  Comm_print("DHCP_SERVER: start DHCP_SERVER\r\n");
  int i;
  for (i=0; i<MAX_ADDRESSES; ++i) { // init addresspool
    addresspool[i].timeout = 0;
    addresspool[i].free = 1;
  }
  Configuration.id = 1;
  Configuration_save();
  // send a package to all clients to collect their addresses
  defaultHeader(&dhcp_packet);
  dhcp_packet.op = 1;
  dhcp_packet.xid = 0;
  dhcp_packet.ciaddr = 0; // current ID
  dhcp_packet.yiaddr = 0;
  dhcp_packet.msgtype = SERVER_DHCPACOLLECT;
  sendDHCPPacket(&dhcp_packet, BROADCAST);
  Comm_print("DHCP_SERVER: DHCPACOLLECT. Collect addresses from clients.\r\n");
  // start the timer which starts the timeout-function 
  Timers_add(TIMEOUT, timeout_f, 0);
  #endif
}

void DHCP_checkHandler() {
  #ifdef DHCPCLIENT
  if (rxPacket.from==Configuration.id) {
    if (dhcp_address_received==1) { // Client
      dhcp_address_received = 0;
      Comm_print("DHCP_CLIENT: Try to get a new address\r\n");
      startDHCP();
    } else if (dhcp_address_received==2) {
      dhcp_address_received = 0;
      startDHCP();
    }
  }
  #endif
}

void DHCP_messageHandler() {
  eDHCP_Header *dhcp_packet;
  eDHCP_Header dhcp_answer;
  // at this point: a DHCP message arrives. Determine what type it is
  dhcp_packet = (eDHCP_Header*)rxPacket.data;
  // The server side
  #ifdef DHCPSERVER
  if (dhcp_packet->msgtype == DHCPDISCOVER) {
    // a client want to have a new address. Search a free one and send it back to the client
    int i;
    Comm_print("DHCP_SERVER: Got DHCPDISCOVER\r\n");
    for (i=0; i<MAX_ADDRESSES; ++i) if (addresspool[i].free == 1) break;
    // i contains the number of the first free address of the addresspool
    if (i==MAX_ADDRESSES) { // no free address available
      Comm_print("DHCP_SERVER: No more free addresses\r\n");
    } else {  // address available
      defaultHeader(&dhcp_answer);
      dhcp_answer.op = 1;   
      dhcp_answer.xid = dhcp_packet->xid;
      dhcp_answer.ciaddr = dhcp_packet->ciaddr;
      dhcp_answer.yiaddr = i+2; // assigned client address
      dhcp_answer.msgtype = DHCPOFFER;
      addresspool[i].free = 0;
      addresspool[i].timeout = 0xFF;
      sendDHCPPacket(&dhcp_answer, dhcp_answer.ciaddr);
      Comm_print("DHCP_SERVER: DHCPOFFER. Assign client address %i\r\n", dhcp_answer.yiaddr);
    }
  } else if (dhcp_packet->msgtype == CLIENT_DHCPACOLLECT) {
  // saves all incoming addresses from clients
  int i;
  Comm_print("DHCP_SERVER: Got DHCPACOLLECT from %i \r\n",dhcp_packet->ciaddr);
  if (dhcp_packet->ciaddr >= MAX_ADDRESSES+2) {
    // address from client is above of maximum number of an available address 
    for (i=0; i<MAX_ADDRESSES; ++i)
    if (addresspool[i].free == 1) {
      defaultHeader(&dhcp_answer);
      dhcp_answer.op = 1;
      dhcp_answer.xid = dhcp_packet->xid;
      dhcp_answer.ciaddr = dhcp_packet->ciaddr;
      dhcp_answer.yiaddr = i+2; // address
      dhcp_answer.msgtype = DHCPOFFER;
      // assign new address
      addresspool[i].free = 0;
      addresspool[i].timeout = 0xFF;
      sendDHCPPacket(&dhcp_answer, dhcp_answer.ciaddr);
      break;
    } else {
      // no free addresses
    }
  } else {
      // address is in range and is set to "not free"
      addresspool[ dhcp_packet->ciaddr-2 ].free = 0;
      addresspool[ dhcp_packet->ciaddr-2 ].timeout = 0xFF;
    }
  } else if (dhcp_packet->msgtype == CLIENT_DHCPREVOKE) {
    // set timeout of client address to initial value
    Comm_print("DHCP_SERVER: Got DHCPREVOKE (timeout). Refresh address %i\r\n", dhcp_packet->ciaddr);
    int i;
    for (i=0; i<MAX_ADDRESSES; ++i)
      if (dhcp_packet->ciaddr == i+2) {
        addresspool[i].timeout = 0xFF;
      }
    }
    #endif
    // The Client Side
    #ifdef DHCPCLIENT
    if (dhcp_packet->msgtype == DHCPOFFER && dhcp_packet->xid==current_xid) {
      // address received from Server; set new ID
      Configuration.id = dhcp_packet->yiaddr;
      Configuration_save();
      dhcp_address_received = 1;
      Comm_print("DHCP_CLIENT: Got DHCPOFFER. New ID is %i.\r\n", Configuration.id);
    } else if (dhcp_packet->msgtype == SERVER_DHCPACOLLECT) {
      // Server has sent this packet to collect client address; 
      // client sends its address
      defaultHeader(&dhcp_answer);
      dhcp_answer.op = 0;   
      dhcp_answer.xid = current_xid; 
      dhcp_answer.ciaddr = Configuration.id;
      dhcp_answer.yiaddr = 0;
      dhcp_answer.msgtype = CLIENT_DHCPACOLLECT;
      sendDHCPPacket(&dhcp_answer, server_address);
      Comm_print("DHCP_CLIENT: Got DHCPACOLLECT. Send address %i to server.\r\n", dhcp_answer.ciaddr);
    } else if (dhcp_packet->msgtype==SERVER_DHCPREVOKE && dhcp_packet->ciaddr==Configuration.id) {
      // timeout runs against 0; send client address to server to refresh timeout
      defaultHeader(&dhcp_answer);
      dhcp_answer.op = 0;
      dhcp_answer.xid = dhcp_packet->xid;
      dhcp_answer.ciaddr = Configuration.id;
      dhcp_answer.yiaddr = 0;
      dhcp_answer.msgtype = CLIENT_DHCPREVOKE;
      sendDHCPPacket(&dhcp_answer, server_address);
      Comm_print("DHCP_CLIENT: Got DHCPREVOKE (timeout). Send address %i to server to refresh.\r\n", Configuration.id);
    }
    #endif
}

bool DHCP_init() {
  #ifdef DHCPSERVER
  dhcp_address_received = 0;
  startDHCP();
  return True;
  #endif
  #ifdef DHCPCLIENT
  server_address = 1;
  dhcp_address_received = 0;
  // generate random MAC-Address for this node
  mac_address = __random32();
  startDHCP();
  return True;
  #endif
}
