/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

/** \file ScatterWeb.DHCP.h
 ** \ingroup Applications
 ** \brief DCHP: Dynamic Host Configuration Protocol (1 server, several clients)
 ** <p>
 ** Introduction 
 ** The server has a fixed address and it manages the address of every client.
 ** TIMEOUT specifies the run of validity.
 ** MAX_ADDRESSES specifies the number of addresses.
 ** <p>
 ** DHCP - packet structure:
 ** <ul>
 ** <li>op - transferred
 ** <li>htype - not transferred, network topology of the Scatterweb is inexistent in RFC 1700
 ** <li>hlen - transferred
 ** <li>hops - transferred, not required in that implementation
 ** <li>xid - transferred
 ** <li>secs - not transferred, needless in that implementation
 ** <li>flags - transferred
 ** <li>ciaddr - transferred, 2 bytes instead of 4 bytes
 ** <li>yiaddr - transferred, 2 bytes instead of 4 bytes
 ** <li>siaddr - transferred, 2 bytes instead of 4 bytes
 ** <li>giaddr - not transferred, no existent sub network
 ** <li>chaddr - transferred, random number for MAC address (2 bytes)
 ** <li>sname - transferred, no entry
 ** <li>file - not transferred, no paths
 ** <li>options - not transferred
 ** </ul>
 ** <p> 
 ** DHCP - messages:
 ** <ul>
 ** <li>DHCPDISCOVER - transferred
 ** <li>DHCPOFFER - transferred
 ** <li>DHCPREQUEST - not transferred, only one server
 ** <li>DHCPACK - not transferred, DHCPOFFER implies DHCPACK
 ** <li>DHCPNAK - not transferred
 ** <li>DHCPRELEASE - not transferred, release the address by timeout
 ** <li>DHCPDECLINE - not transferred, if duplicate address identified -> get new address by server
 ** <li>DHCPINFORM - not transferred, inexistent configuration parameter
 ** </ul>
 ** <p> 
 ** DHCP - additional messages:
 ** <ul>
 ** <li>SERVER_DHCPREVOKE - demand of the server on the client: refresh your address
 ** <li>CLIENT_DHCPREVOKE: sending the address, which has to refresh (in reply to SERVER_DHCPREVOKE)
 ** <li>SERVER_DHCPACOLLECT - demand of the server on all clients: send back your addresses
 ** <li>CLIENT_DHCPACOLLECT: sending the address (in reply to SERVER_DHCPACOLLECT)
 ** </ul>
 ** <p>
 ** Add-on
 ** Clients are able to detect that another node use the same address
 ** and will be relocated (server-side).
 **
 ** @author        Marco Hutta
 ** @version       2.2
 ** @since         2.2
 **/ 
 
#ifndef ScatterWeb_DHCP
#define ScatterWeb_DHCP

#include "../../../../System/src/ScatterWeb.System.h"

/**
 * DHCPSERVER or DHCPCLIENT
 *
 * @since         2.2
 */
//#define DHCPSERVER
#define DHCPCLIENT

/**
 * Maximum number of addresses
 *
 * @since         2.2
 */
#define MAX_ADDRESSES     100

/**
 * Time to expire of address
 *
 * @since         2.2
 */
#define TIMEOUT            70

/**
 * initiates and starts the DHCP Server/Client
 *
 * @return        <code>True</code> always. 
 * @since         2.2
 */
bool DHCP_init();

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
void DHCP_messageHandler();

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
void DHCP_checkHandler();

#endif

