/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

/** \file ScatterWeb.Comm.h
 ** \ingroup System
 ** \brief The ScatterWeb.Comm represents the standard output for applications.
 ** <p>
 ** The init function for the uart Comm_init() provides setup for all standard
 ** baud rates: BPS_2400, BPS_14400, BPS_19200, BPS_38400, BPS_57600, and
 ** BPS_115200.
 ** <p>
 ** All uart communication should be done with Comm_print() or Comm_log() as the
 ** serial output with this functions is controlled via Com_logLevel and struct
 ** Configuration.boot_logLevel in ScatterWeb.Configuration. All sending
 ** functions are blocking. 
 ** <p>
 ** Receiving is handled by an interrupt and you can choose byte-access or
 ** line-access.
 ** <p>
 ** byte-access:
 ** To gain access to the byte level, register a function that expect to get
 ** pass an UINT8 as parameter (see Comm_registerByteLevel(fp_char_t fp)).
 ** <p>
 ** line-access:
 ** To gain access to the line level, register a function that expect to get
 ** pass an void* as parameter (see C_SERIAL and System_registerCallback(UINT8
 ** type, void(*fp_t)() )). If the line is handled the receiving function should
 ** call Comm_clear().
 **
 ** @author        Marco Hutta
 ** @author        Bjoern Lichtblau
 ** @version       2.3
 ** @since         2.0
 **/  
 
#ifndef ScatterWeb_Comm
#define ScatterWeb_Comm

/**
 * Define that sets the baudrate to 2400 Bit/s
 *
 * @see           #Comm_init(UINT8 bps)
 * @since         2.0
 */
#define BPS_2400       (0x01)

/**
 * Define that sets the baudrate to 14400 Bit/s
 *
 * @see           #Comm_init(UINT8 bps)
 * @since         2.0
 */
#define BPS_14400      (0x02)

/**
 * Define that sets the baudrate to 19200 Bit/s
 *
 * @see           #Comm_init(UINT8 bps)
 * @since         2.0
 */
#define BPS_19200      (0x03)

/**
 * Define that sets the baudrate to 38400 Bit/s
 *
 * @see           #Comm_init(UINT8 bps)
 * @since         2.0
 */
#define BPS_38400      (0x04)

/**
 * Define that sets the baudrate to 57600 Bit/s
 *
 * @see           #Comm_init(UINT8 bps)
 * @since         2.0
 */
#define BPS_57600      (0x05)

/**
 * Define that sets the baudrate to 115200 Bit/s
 *
 * @see           #Comm_init(UINT8 bps)
 * @since         2.0
 */
#define BPS_115200     (0x06)

/**
 * Value of the log level
 *
 * @see           #Comm_log(UINT8, const char*, ...)
 * @since         2.0
 */
UINT8 Comm_logLevel;     

/**
 * Initializes the serial
 *
 * @param bps     the bps to be set 
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool Comm_init(UINT8 bps);

/**
 * Sends bytes of data
 *
 * @param fmt     function
 * @param ...     list of parameters
 * @return        <code>True</code> if serial is on;
 *                <code>False</code> otherwise.
 * @since         2.0
 */
bool Comm_print(const char *fmt, ...);

/**
 * Sends bytes of log data
 *
 * @param level   log level
 * @param fmt     function
 * @param ...     list of parameters 
 * @return        <code>True</code> if serial is on;
 *                <code>False</code> otherwise.
 * @since         2.0
 */
bool Comm_log(UINT8 level, const char* fmt, ...);

/**
 * Switchs the serial on
 *
 * @return        <code>True</code> always.
 * @since         2.2
 */
bool Comm_on();

/**
 * Switchs the serial off
 *
 * @return        <code>True</code> always.
 * @since         2.2
 */
bool Comm_off();

/**
 * Registers byte level and will be set by Comm_init(UINT8 bps) 
 *
 * @param fp      pointer of function 
 * @return        <code>True</code> if ECR or ESB defined
 *                <code>False</code> if eGate defined.
 * @see           #Comm_init(UINT8 bps) 
 * @since         2.2
 */
bool Comm_registerByteLevel(fp_char_t fp);

/**
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Comm_clear();  
  
#endif
