/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

/** \file ScatterWeb.Configuration.h
 ** \ingroup System
 ** \brief The ScatterWeb.Configuration provides fuctions that allow you to
 ** programmatically access configuration settings.
 ** <p>
 ** The configuration consists of the struct config_t defined in
 ** ScatterWeb.Configuration.h. The variable Configuration is of that type and
 ** represents the current config in RAM and can be accessed directly everytime.
 ** <p>
 ** Changes to Configuration will only be made permanent (power-off & resets)
 ** with a call to Configuration_save(), which copies the content of the 
 ** Configuration to EEPROM[EADR_CONFIG].
 ** <p>
 ** Initial loading from EEPROM[EADR_CONFIG] is done by Configuration_init() and
 ** has to be called on initialization in main().
 ** <p>
 ** First time configuration:
 ** Uncomment the following definition of firstTimeConfig and it will get layed
 ** into Infomem (0x1000-0x1100, see .hex-File). Before Configuration_init()
 ** loads the regular config from EEPROM, it checks if the first int (which
 ** corresponds to the id) in Infomem is not 0xFFFF. If it is not, it copies the
 ** configuration from Infomem to EEPROM (clearing Infomem afterwards and
 ** loading the newly written data from EEPROM to Configuration).
 **
 ** @author        Marco Hutta
 ** @author        Bjoern Lichtblau
 ** @author        Thomas Pietsch 
 ** @version       2.3
 ** @since         2.0
 **/  
 
#ifndef ScatterWeb_Configuration
#define ScatterWeb_Configuration

/**
 * Base address in EEPROM for configuratio data.
 *
 * @since         2.0
 */
#define EADR_CONFIG     (0x0F00)

/**
 * Radio debug output.
 *
 * @see           #Comm_log(UINT8, const char*, ...) 
 * @see           #Configuration.boot_logLevel 
 * @since         2.0
 */
#define DEBUG_RADIO               (0x80)

/**
 * Dco adjusting debug output.
 *
 * @see           #Comm_log(UINT8, const char*, ...) 
 * @see           #Configuration.boot_logLevel 
 * @since         2.0
 */
#define DEBUG_DCO                 (0x40)

/**
 * Synchronisation debug output.
 *
 * @see           #Comm_log(UINT8, const char*, ...) 
 * @see           #Configuration.boot_logLevel 
 * @since         2.2
 */
#define DEBUG_SYNC                (0x20)

/**
 * Comm_print() gets through.
 *
 * @see           #Comm_log(UINT8, const char*, ...) 
 * @see           #Configuration.boot_logLevel 
 * @since         2.0
 */
#define NO                        (0x01)

/**
 * Comm_print(), Comm_log(LOW) get through.
 *
 * @see           #Comm_log(UINT8, const char*, ...) 
 * @see           #Configuration.boot_logLevel 
 * @since         2.0
 */
#define LOW                       (0x02)

/**
 * Comm_print(), Comm_log(LOW), Comm_log(MEDIUM) get through.
 *
 * @see           #Comm_log(UINT8, const char*, ...) 
 * @see           #Configuration.boot_logLevel 
 * @since         2.0
 */
#define MEDIUM                    (0x03)

/**
 * Comm_print(), Comm_log(LOW), Comm_log(MEDIUM), Comm_log(HIGH) get through.
 *
 * @see           #Comm_log(UINT8, const char*, ...) 
 * @see           #Configuration.boot_logLevel 
 * @since         2.0
 */
#define HIGH                      (0x04)

/**
 * All get through.
 *
 * @see           #Comm_log(UINT8, const char*, ...) 
 * @see           #Configuration.boot_logLevel 
 * @since         2.0
 */
#define VERBOSE                   (0x05)

/**
 * DPanic if wd_stackCheck detects a low stack.
 *
 * @see           #Configuration.panicFlags
 * @since         2.0
 */
#define PANIC_ON_STACK_OVERFLOW   (0x01)

/**
 * Panic if timers get full.
 *
 * @see           #Configuration.panicFlags
 * @since         2.0
 */
#define PANIC_ON_FULL_TIMERS      (0x02)

/**
 * Panic if last reset was a bad reset.
 *
 * @see           #Configuration.panicFlags
 * @since         2.0
 */
#define PANIC_ON_BAD_RESET        (0x04)

/**
 * If set and a Panic occured, dont initialize the application.
 *
 * @see           #Configuration.panicFlags
 * @since         2.0
 */
#define PANIC_DISABLES_APP        (0x80)

/**
 * If set device allows reprogramming itself.
 *
 * @see           #Configuration.firmwareFlags 
 * @since         2.0
 */
#define FF_PROGRAMMABLE            (0x01)
/**
 * If set dco adjusting will be enabled.
 *
 * @see           #Configuration.firmwareFlags 
 * @since         2.0
 */
#define FF_DCOCHECKER              (0x02)

/**
 * Configuration of the device
 * <ul>
 * <li>id - Id of this MSP device
 * <li>rxReceiveLimit - Threshold for receive power to realize a "Carrier Detect" with ADU.
 * <li>transmitPower - Power for transmittting, set via Net_setTxPower at boot time.
 * <li>boot_logLevel - Comm_logLevel is initialized with that value.
 * <li>panicFlags - Controls panic behaviour.
 * <li>firmwareFlags - Other bits for firmware behaviour.
 * </ul>
 *  
 * @since         2.0
 */
typedef struct {
  UINT16 id;
  UINT16 rxReceiveLimit;
  UINT8 transmitPower;
  UINT8 boot_logLevel;
  UINT8 panicFlags;
  UINT8 firmwareFlags;
} config_t;

/**
 * Copy of the Configuration
 *
 * @since         2.0
 */
extern config_t Configuration;

/**
 * Initializes configuration settings for the configuration section
 *
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool Configuration_init();

/**
 * Saves configuration settings for the configuration section
 *
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool Configuration_save();

/**
 * Prints configuration settings for the configuration section
 *
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool Configuration_print(); 

#endif
