/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

/** \file ScatterWeb.Data.h
 ** \ingroup System
 ** \brief The ScatterWeb.Data consists mostly of the fuctions that constitute
 ** the sensors architecture. The sensors architecture enables you to build
 ** components that efficiently manage data from multiple data sources. It
 ** provides the tools to initialize, enable, and disable data in this multiple
 ** tier system.
 ** <p>
 ** To get sensor events you must set a callback function in 
 ** System_callbacks[C_SENSOR], for disabling that set 
 ** System_callbacks[C_SENSOR] = NULL.   
 ** <p>
 ** Each sensor is seperately enabled and disabled with its enable/disable
 ** function. The enable function takes some sensor specific arguments. The
 ** first argument is the interval (or debounce delay) for reading the sensor
 ** internally, it can also be seen as a reaction time (for example for Vib: If
 ** Vibration has been signaled how long must the sensor stay silent to signal
 ** NoVibration? ).
 ** <p>
 ** For temperature 4 arguments are needed to specify a funnel function: 
 ** Data_temperatureOn(Interval, StartRange, Diff, StepsPerDiff) 
 ** <ul>
 ** <li>Interval - How often is the temperature read in systems ticks (~1ms).
 ** <li>StartRange - How much must the temperature differ from the last value to get signaled.
 ** <li>Diff - How much the Range is decreased every DiffStep measures.
 ** <li>StepsPerDiff - How much measures until Range is decremented with Diff.
 ** </ul>
 ** <p>
 ** Attention: Temperature data is in fixed point 2s complement format!!! 
 ** <p>
 ** This approach also includes simple standard evaluation methodes:
 ** - Periodic Reading: To signal every x ms do Data_temperatureOn(x, 0, 0, 0);
 ** - On change Reading: To signal only on temperature changes greater than x
 **     (temperature is read every y ms) do Data_temperatureOn(y, x, 0, 0):  
 ** - Periodic & On Change combined: To signal on every change greater than 
 **     x or after a time y*z has passed (temperature is read every z ms) do
 **     Data_temperatureOn(z, x, x, y).
 ** <p>
 ** Examples: 
 ** <ul>
 ** <li>Data_temperatureOn(5120, 0, 0, 0) - Every 5 seconds.
 ** <li>Data_temperatureOn(2048, 0x0080, 0, 0) - Only on change greater 1C (temperature is read every 2 seconds).
 ** <li>Data_temperatureOn(6144, 0x0080, 0x0080, 20)- Every 20*6s=2min or if changed more than 1C (temperature is read every 6 seconds).
 ** <li>Data_temperatureOn(6144, 0x0200, 0x0040, 10) - Funnel function: Measure every 6 seconds and decrease range from 2C for 0,5C every 10*6s=1min (temperature is read every 6 seconds).
 ** </ul>
 ** <p>
 ** Power for sensors (and RS232 voltage converter) must be enabled, see
 ** System_powerOn() / System_powerOff().
 ** <p>
 ** The IR Receiver is handled in file ScatterWeb.Data.c.
 ** <p>
 ** A good example for using the sensors is ScatterWeb.Event.c in [EMPTY]\ESB
 ** project. 
 ** <p>
 ** Controlling the beeper.
 ** <p>
 ** Receiving RC5 via IR Receiving Diode.
 ** <p>
 ** RC5: 1780 us bitlength (manchester encoded, so half bitlength of 890 us is
 ** important). Transferred packet (2 start + toggle bit + 5 address bits + 6
 ** comand bits)): 
 **                                   | S | S | T | A4 | A3 | A2 | A1 | A0 | C5 | C4 | C3 | C2 | C1 | C0 |
 ** irdata format: | ? | ? | error  | newData | T | A4 | A3 | A2 | A1 | A0 | C5 | C4 | C3 | C2 | C1 | C0 |
 ** <p>
 ** Some common addresses and commands:
 ** <p>
 ** Address:          Device:          Command:
 **   0               TV1              0...9    Numbers 0...9 (channel select)
 **   1               TV2              12       Standby
 **   5               VCR1             16       Master Volume +
 **   6               VCR2             17       Master Volume -
 **  17               Tuner            18       Brightness +
 **  18               Audio Tape       19       Brightness -
 **  20               CD Player        50       Fast rewind
 **                                    52       Fast run forward
 **                                    53       Play
 **                                    54       Stop
 **                                    55       Recording
 ** <p>
 ** Methods for sending with ir diode.
 ** <p>
 ** These methods are timed to the system clock (4505600 Mhz) and disable
 ** all interrupts and so will halt the system timer for some milliseconds.
 ** Radio packet transmission & reception are also disrupted. 
 ** 
 ** @author          Marco Hutta
 ** @author          Bjoern Lichtblau
 ** @version         2.3
 ** @since           2.0  
 **/

#ifndef ScatterWeb_Data
#define ScatterWeb_Data

#include "ScatterWeb.Time.h"

/**
 * movement sensor (on P13)
 *
 * @see             #sensorFlags
 * @since           2.0
 */
#define SENSOR_MOVEMENT     (0x08)

/**
 * vibration sensor (on P14)
 *
 * @see             #sensorFlags
 * @since           2.0
 */
#define SENSOR_VIBRATION    (0x10)

/**
 * button (on P27)
 *
 * @see             #sensorFlags
 * @since           2.0
 */
#define SENSOR_BUTTON       (0x80)

/**
 * microphone (on P60).
 *
 * @see             #sensorFlags
 * @since           2.0
 */
#define SENSOR_MICROPHONE   (0x01)

/**
 * temperature sensor (DS1629 or LM71)
 *
 * @see             #sensorFlags
 * @since           2.0
 */
#define SENSOR_TEMPERATURE  (0x02)

/**
 * RC5 receiving via IR diode
 *
 * @see             #sensorFlags
 * @since           2.0
 */
#define SENSOR_RC5          (0x04)

/**
 * supply voltage
 *
 * @see             #sensorFlags
 * @since           2.0
 */
#define SENSOR_VOLTAGE      (0x40)

/**
 * Holds sensor data record.
 * <ul>
 * <li>id - Id of node these readings belong to.
 * <li>sensor - Must be one of SENSOR_MOVEMENT, SENSOR_VIBRATION, SENSOR_BUTTON, SENSOR_MICROPHONE, SENSOR_TMPERATURE, SENSOR_RC5, SENSOR_VOLTAGE.
 * <li>timeStamp - Timestamp of sensor reading.
 * <li>value - Value of the reading.
 * </ul>
 *  
 * @since           2.0
 */
typedef struct {
  UINT16 id;
  UINT8 sensor;
  time_t timeStamp;
  UINT16 value;
} sdata_t;

/**
 * Value of SensorFlags
 *
 * @since           2.0
 */
extern UINT8 sensorFlags;

/**
 * Initializes the sensors
 *
 * @return          <code>True</code> if sensors available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_init();

/**
 * Enables the battery sensor
 *
 * @param interval  interval of measurement
 * @param thresh    thresh of measurement
 * @return          <code>True</code> if battery sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_batteryOn(UINT16 interval, UINT16 thresh);

/**
 * Disables the battery sensor
 *
 * @return          <code>True</code> if battery sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_batteryOff();

/**
 * Enables the beeper
 *
 * @return          <code>True</code> if beeper sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_beeperOn();

/**
 * Disables the beeper
 *
 * @return          <code>True</code> if beeper sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_beeperOff();

/**
 * Toggles the beeper
 *
 * @return          <code>True</code> if beeper sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_beeperToggle();

/**
 * Enables the button
 *
 * @param delay     delay of measurement
 * @return          <code>True</code> if button sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_buttonOn(UINT16 delay);

/**
 * Disables the button
 *
 * @return          <code>True</code> if button sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_buttonOff();

/**
 * Enables the green LED
 *
 * @return          <code>True</code> if green led available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_greenOn();

/**
 * Disables the green LED
 *
 * @return          <code>True</code> if green led available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_greenOff();

/**
 * Toggles the green LED
 *
 * @return          <code>True</code> if green led available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_greenToggle();

/**
 * Enables the red LED
 *
 * @return          <code>True</code> if red led available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_redOn();

/**
 * Disables the red LED
 *
 * @return          <code>True</code> if red led available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_redOff();

/**
 * Toggles the red LED
 *
 * @return          <code>True</code> if red led available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_redToggle();

/**
 * Enables the yellow LED
 *
 * @return          <code>True</code> if yellow led available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_yellowOn();

/**
 * Disables the yellow LED
 *
 * @return          <code>True</code> if yellow led available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_yellowOff();

/**
 * Toggles the yellow LED
 *
 * @return          <code>True</code> if yellow led available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_yellowToggle();

/**
 * Enables the microphone sensor
 *
 * @param delay     delay of measurement
 * @param diff      diff of measurement
 * @return          <code>True</code> if microphone sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_microphoneOn(UINT16 delay, UINT16 diff);

/**
 * Disables the microphone sensor
 *
 * @return          <code>True</code> if microphone sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_microphoneOff();

/**
 * Enables the movement sensor
 *
 * @param delay     delay of measurement
 * @return          <code>True</code> if movement sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_movementOn(UINT16 delay);

/**
 * Disables the movement sensor
 *
 * @return          <code>True</code> if movement sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_movementOff();

/**
 * Enables the temperature sensor
 *
 * @param interval  interval of measurement
 * @param range     range of measurement
 * @param diff      diff of measurement
 * @param spd       spd of measurement 
 * @return          <code>True</code> if temperature sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_temperatureOn(UINT16 interval, UINT16 range, UINT16 diff, UINT16 spd);

/**
 * Disables the temperature sensor
 *
 * @return          <code>True</code> if temperature sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_temperatureOff();

/**
 * Enables the vibration sensor
 *
 * @param delay     delay of measurement
 * @return          <code>True</code> if vibration sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_vibrationOn(UINT16 delay);

/**
 * Disables the vibration sensor
 *
 * @return          <code>True</code> if vibration sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */ 
bool Data_vibrationOff();

/**
 * Returns the state of the green LED
 *
 * @return          <code>True</code> if green Led is switched on;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_greenLed();

/**
 * Returns the state of the red LED
 *
 * @return          <code>True</code> if red Led is switched on;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_redLed();

/**
 * Returns the state of the yellow LED
 *
 * @return          <code>True</code> if yellow Led is switched on;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_yellowLed();

/**
 * Sends a RC5 PHI command
 *
 * @param data      data to send
 * @return          <code>True</code> True if ir sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_RC5SendPHI(UINT16 data);

/**
 * Enables the RC5
 *
 * @return          <code>True</code> True if ir sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_RC5Enable();

/**
 * Disables the RC5
 *
 * @return          <code>True</code> True if ir sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_RC5Disable();

/**
 * Sends a RC5 NEC command
 *
 * @param data1     data1 to send
 * @param data2     data2 to send
 * @return          <code>True</code> True if ir sensor available;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Data_RC5SendNEC(UINT16 data1, UINT16 data2);

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Data_sensorHandler();

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
void Data_RC5ReceiveHandler();

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
void Data_RC5EventHandler();  

#endif
