/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

/** \file ScatterWeb.IO.h
 ** \ingroup System
 ** \brief The ScatterWeb.IO contains functions that allow reading and writing
 ** to the EEPROM.
 ** <p>
 ** Most important functions for writing/reading the EEPROM are IO_write() /
 ** IO_read().
 ** <p>
 ** The EEPROM offers 64k memory (100000 write cycles) and is accessed with an
 ** I2C like bus.
 ** <p>
 ** Currently use is storing configuration data (see ScatterWeb.Configuration.c) 
 ** and temporary storage of a new software image before it gets flashed.
 ** <p>
 ** ( DEVICE ADDRESSING:   Start | 1 0 1 0 A2 A1 A0 R/W | ... )
 ** <p>
 ** EEPROM memory map:
 ** <ul>
 ** <li>Region - Usage
 ** <li>0x0000 - 0x0CFF - Free for use.
 ** <li>0x0D00 - 0x0DFF - Application config (EADR_APPCONFIG).
 ** <li>0x0E00 - 0x0EFF - Used by flasher.c / EADR_FLASHERBUF: Free for use only if flasher_state==IDLE
 ** <li>0x0F00 - 0x0FFF - Never change something here. Used for configuration data (ScatterWeb.Configuration.c, EADR_CONFIG) and some reprogramming variables.
 ** <li>0x1000 - 0xFFFF - Programming image storage or other big data. Use according to EEPROM[EADR_IMAGESTATE].
 ** </ul>
 **
 ** @author          Marco Hutta
 ** @author          Bjoern Lichtblau
 ** @author          Thomas Pietsch 
 ** @version         2.3
 ** @since           2.0   
 **/

#ifndef ScatterWeb_IO
#define ScatterWeb_IO

/**
 * Temporary switch for the eeprom address. Use 8k EEPROM, future versions will
 * have only one 64k eeprom at this address.
 *
 * @since           2.0
 */
#define EEPROMADDRESS (0x00)

/**
 * Must be set together with EEPROMADDRESS. 64k EEPROM, 128b page writes
 * possible.
 *
 * @since           2.0
 */
#define EEPROMPAGEMASK (0x7F)

/**
 * Erase several bytes from eeprom (means set them to 0xFF).
 *
 * @param e_adr     start address of bytes to erase
 * @param size      size of bytes to erase
 * @return          <code>True</code> if erase successful;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool IO_erase(UINT16 e_adr, UINT16 size);

/**
 * Reads size bytes from the eeprom, uses sequential read (faster than
 * readEEPROMByte).
 *
 * @param e_adr     start address of bytes to read
 * @param r_adr     array to store read bytes
 * @param size      size of bytes to read
 * @return UINT8*   pointer of r_adr
 * @since           2.0
 */
UINT8* IO_read(int e_adr, char* r_adr, int size);                                                                         

/**
 * Writes size bytes to the eeprom, uses sequential write (faster than
 * writeEEPROMByte).
 *
 * @param e_adr     start address of bytes to write
 * @param r_adr     array of bytes to write
 * @param size      size of bytes to write
 * @return          <code>True</code> if write successful;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool IO_write(UINT16 e_adr, UINT8* r_adr, UINT16 size);

#endif
