/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

/** \file ScatterWeb.Net.h
 ** \ingroup System
 ** \brief The ScatterWeb.Net implements a network service that enables you to
 ** develop applications that use network resources without worrying about the
 ** specific details of the individual protocol.
 ** <p>
 ** This file contains following functionality for radio:
 ** <p>
 ** - Setting of transmit power (in 100 steps).
 ** - Interrupt driven Rx&Tx radio MAC logic (sending of packets with CSMA and
 **   random backoff, CRC-16 checksums, acknowledgement with automatic
 **   retransmission).
 ** <p>
 ** General
 ** Radio software supports acknowledgements, so it is important to know that
 ** all packet types which are no broadcast will get acknowledged by radio
 ** software and get repeated several times if needed. The radio software uses
 ** the num field for acknowledged packets to recognize doubles. 
 ** <p>
 ** packet_t is a struct containing a packet header fields and pointers to
 ** packet data and is used for sending & receiving. 
 ** <p>
 ** As a packet may be often constructed out of a header portion and a data
 ** portion, packet_t supports two data pointers to get passed in for sending,
 ** so these portions don't need to be copied together. When receiving however,
 ** data pointer contains header and data consecutively and header and
 ** header_length are null.
 ** <p>
 ** How to receive packets
 ** To receive packets of a specific type one must define a handler function and
 ** add it to Net_txHandler for firmware handlers or to the C_RADIO app callback
 ** function. 
 ** <p>
 ** When the handler is then called, the function can access the received packet
 ** in rxPacket. 
 ** <p>
 ** How to send packets
 ** For sending packets a packet_t must be initialized:
 ** - to must contain the destination id of receiver or broadcast address.
 ** - type must contain the packet type.
 ** - header/data must be set to NULL or to the data that should get
 **   transported. If set, header_length/data_length should contain the length
 **   of the data.
 ** - num can be set to transport a byte for not acknowledged packets, otherwise
 **   it will get overwritten.
 ** - from is always overwritten with the id of this device.
 ** <p>
 ** Such a packet can then be queued into tx ring buffer with Net_send and a
 ** pointer to initialized packet_t.
 ** <p>
 ** Additionally these two functions accept a callback pointer to functions
 ** which accept an UINT8. If set, this callback is called when the packet is
 ** removed from the tx ring buffer:
 ** - For unacknowledged packets this is after sending the packet.
 ** - For acknlowedged packets this is after successfully receiving an ACK for
 **   this packet (callback argument is set to 1) or when all retries are made
 ** without sucess (callback argument is set to 0). 
 ** <p>
 ** The callback function can check the variable txRetries to see how much 
 ** retries it took (always 0 for unacknowledged packets).
 ** 
 ** @author        Marco Hutta
 ** @author        Bjoern Lichtblau
 ** @version       2.3
 ** @since         2.0     
 **/

#ifndef ScatterWeb_Net
#define ScatterWeb_Net

#include "ScatterWeb.System.h"

/**
 * Predefined Address: Broadcast Address.
 *
 * @since         2.0
 */
#define BROADCAST             (0xFFFF)

/**
 * Predefined Address: Illegal Address.
 *
 * @since         2.0
 */
#define UNKNOWN               (0x0000)

/**
 * Packet Type: acknowledgment packet.
 *
 * @since         2.0
 */
#define ACK_PACKET            (0x00)

/**
 * Packet Type: sync packet.
 *
 * @since         2.0
 */
#define SYNC_PACKET          (0x01)

/**
 * Packet Type: Simple ping.
 *
 * @since         2.0
 */
#define PING_PACKET           (0x02)

/**
 * Packet Type: Simple pong.
 *
 * @since         2.0
 */
#define PONG_PACKET           (0x03)

/**
 * Packet Type: Module Messaging - Terminal request containing the terminal
 * command.
 *
 * @since         2.0
 */
#define MESSAGING_REQ_PACKET  (0x04)

/**
 * Packet Type: Module Messaging - Terminal answer.
 *
 * @since         2.0
 */
#define MESSAGING_ANS_PACKET  (0x05)

/**
 * Packet Type: Application Scan-Net - Request for SCAN_REPLY.
 *
 * @since         2.0
 */
#define SCAN_PACKET           (0x06)

/**
 * Packet Type: Application Scan-Net - Answer with information about device type
 * and current software.
 *
 * @since         2.0
 */
#define SCAN_REPLY_PACKET     (0x07)

/**
 * Packet Type: Packet for sensor values.
 *
 * @since         2.0
 */
#define SENSOR_PACKET         (0x08)

/**
 * Packet Type: Library Flasher - Initial request to a target to accept a new image.
 *
 * @since         2.1
 */
#define FLASH_HEADER          (0x11)

/**
 * Packet Type: Library Flasher - Acknowledge for HEADER (and FIN for multiple DEVICES),
 * Deny if busy, not programmable or other error.
 *
 * @since         2.1
 */
#define FLASH_HEADER_REPLY    (0x12)

/**
 * Packet Type: Library Flasher - Software image data chunk.
 *
 * @since         2.1
 */
#define FLASH_DATA            (0x14)

/**
 * Packet Type: Library Flasher - End of transmission packet. Acknowledge to
 * FIN. Request if a data chunk is still missing to FIN.
 *
 * @since         2.1
 */
#define FLASH_REQ             (0x15)

/**
 * Packet Type: Application PictureIt! - Requesting a Picture.
 *
 * @since         2.0
 */
#define PICTURE_REQ_PACKET    (0x16)

/**
 * Packet Type: Application PictureIt! - Answer to PICTURE_REQ (containing
 * failure message or picture data).
 *
 * @since         2.0
 */
#define PICTURE_ANS_PACKET    (0x17)

/**
 * Packet Type: Application ScatterRouting - packet for routing state.
 *
 * @since         2.1
 */
#define ROUTING_INFO          (0x20)

/**
 * Packet Type: Application ScatterRouting - Request childhood at a new parent.
 *
 * @since         2.1
 */
#define ROUTING_REQUEST       (0x22)

/**
 * Packet Type: Application ScatterRouting - Accept a new child.
 *
 * @since         2.1
 */
#define ROUTING_ACCEPT        (0x23)

/**
 * Packet Type: Application ScatterRouting - Deny a new child.
 *
 * @since         2.1
 */
#define ROUTING_DENY          (0x24)

/**
 * Packet Type: Application ScatterRouting - To inform a parent of a wrong childhood.
 *
 * @since         2.1
 */
#define ROUTING_NOCHILD       (0x25)

/**
 * Packet Type: Application ScatterRouting - Packets destinated to the root.
 *
 * @since         2.1
 */
#define ROUTING_ROOTPACKET    (0x26)

/**
 * Packet Type: Application ScatterRouting - Packets flooded through the tree by
 * the root.
 *
 * @since         2.1
 */
#define ROUTING_TREEPACKET    (0x27)

/**
 * Packet Type: Application RSA - Requesting the Public key from communication
 * partner
 *
 * @since         2.2
 */
#define RSA_KEYREQUEST        (0x28)

/**
 * Packet Type: Application RSA - Replying Public key to the communication
 * partner
 *
 * @since         2.2
 */
#define RSA_KEYREPLY          (0x29)

/**
 * Packet Type: Application eDHCP
 *
 * @since         2.2
 */
#define DHCP_PACKET           (0x30)

/**
 * Packet Type: Application vGate
 *
 * @since         2.3
 */
#define VGATE_REPLY           (0x31)

/**
 * Packet Type: Application DirectedDiffusion
 *
 * @since         2.3
 */
#define DD_INTEREST_MESSAGE   (0x32)

/**
 * Packet Type: Application DirectedDiffusion
 *
 * @since         2.3
 */
#define DD_MGMT_PACKET        (0x33)

/**
 * Packet Type: Application DirectedDiffusion
 *
 * @since         2.3
 */
#define DD_DATA_MESSAGE       (0x34)

/**
 * Packet Type: Application DirectedDiffusion
 *
 * @since         2.3
 */
#define DD_UPDATE_MESSAGE     (0x35)

/**
 * Packet Type: Application DirectedDiffusion
 *
 * @since         2.3
 */
#define DD_ROUTING_MESSAGE    (0x36)

/**
 * Packet Type: Application DirectedDiffusion
 *
 * @since         2.3
 */
#define DD_ROUTABLE_PACKET    (0x37)

/**
 * Currently receiving?
 *
 * @since         2.0
 */
#define IS_RADIO_RECEIVING (cdCounter)

/**
 * Switched to TX?
 *
 * @since         2.0
 */
#define IS_RADIO_SENDING   ((P5OUT & 0xC0)==0x40)

/**
 * Struct for the packet
 * <ul>
 * <li>to - Destination.
 * <li>from - Source (automatically set by radioSend).
 * <li>type - Packet type.
 * <li>num - Packet number.
 * <li>header - Pointer to packet extra header.
 * <li>header_length - 0 means no extra header (in radioReceive always 0)
 * <li>data - Pointer to packet data.
 * <li>data_length - Length of data.
 * </ul>
 *  
 * @since         2.0
 */
typedef struct {
  UINT16 to;
  UINT16 from;
  UINT8 type;
  UINT8 num;
  UINT8* header;
  UINT16 header_length;
  UINT8* data;
  UINT16 data_length;
} packet_t;

/**
 * struct for received packet
 *
 * @since         2.0
 */
extern volatile packet_t rxPacket;

/**
 * struct for sended packet
 *
 * @since         2.0
 */
extern packet_t sentPacket;

/**
 * number of retries done (retries needed to deliver a packet).
 *
 * @since         2.0
 */
extern volatile UINT8 txRetries;

/**
 * indicate if radio free or not
 *
 * @since         2.0
 */
extern volatile UINT8 cdCounter;

/**
 * state of receiving state machine (public for debugging).
 *
 * @since         2.0
 */
extern volatile UINT8 txState;

/**
 * Initializes the radio
 *
 * @return          <code>True</code> always;
 * @since           2.0
 */
bool Net_init();

/**
 * Enables the radio
 *
 * @return          <code>True</code> always;
 * @since           2.0
 */
bool Net_on();

/**
 * Disables the radio
 *
 * @return          <code>True</code> always;
 * @since           2.0
 */
bool Net_off();

/**
 * Sets the transceiver power
 *
 * @param value     tx power to set
 * @return          <code>True</code> if value are between 0 and 100;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Net_setTxPower(UINT8 value);

/**
 * Sends a packet over radio
 *
 * @param packet    packet to send
 * @param callback  will call after sending (without: use NULL)
 * @return          <code>True</code> if packet is successfully buffered;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Net_send(packet_t* packet, fp_char_t callback);

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Net_rxHandler();

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Net_txHandler();

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
char Net_esend(packet_t* packet, fp_char_t callback);

#endif
