/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

#include "ScatterWeb.System.h"

#define minuteDuration        60000 // how long should our "virtual minute" take? (in msecs)
#define secsAwakeBeforeSync   35    // seconds to wake up before our software clock tells us a synch message will
                                    // be received again, necessary for clock drifts to avoid missing sync messages

UINT8 isRoot;
UINT8 isInit;
UINT16 arrivalTimeMillis; // software time, when sync packet arrived or - if root - when synch message
UINT32 arrivalTimeSecs;   // was sent
UINT32 wakeUpTime = 0;

extern volatile time_t currentTime;
extern volatile time_t rxTime;
extern volatile packet_t rxPacket;  

fp_t callbackSyncDone;

// switch on
void startSync() {
  Net_on();
  if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Synchronisation begins at: %i secs\r\n",currentTime.secs);
}

void waitForBeginSync() {
  // if time to wake up has come...
  if (currentTime.secs > wakeUpTime) startSync();
  else {
    // how long we have to wait
    UINT32 timeRemaining = wakeUpTime - currentTime.secs;
    // this is needed cause function does not accept 'big' numbers (see above)
    if (timeRemaining > 60) {
        // come back in a minute
        Timers_add(60000, waitForBeginSync, 0xFFFF); 
    }
    // we have less than one minute to wait, so come back soon
    else Timers_add(timeRemaining, waitForBeginSync, 0); 
  }
}

// switch off
void endSync() {
  if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Synchronisation ends at: %i secs\r\n",currentTime.secs);
  if(callbackSyncDone) callbackSyncDone();
  waitForBeginSync();
}

// queues a broadcast sync message, in send function the source will be the master
// root id, and num will be the seqNum
void sendSync(UINT16 minutes) {
  packet_t packet;
  time_t t = {0, 0};
  // enqueue this function again, if not minutes have passed
  // so for send frequency of e.g. 4 minutes this function gets
  // enqueue 4 times
  // this is done because the first argument of timer_add will be
  // wrapped if too big millis
  if (minutes > 0) { // minutes havn't passed so reput in timer and don't send
    if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Still %i minute(s) to go\r\n", minutes);
    Timers_add(minuteDuration, sendSync, --minutes); 
    return;
  }
  // check if the last received message from the root
  // is too old
  if (SyncStruct.syncPeriods >= ROOT_TIMEOUT && SyncStruct.masterRootId != Configuration.id) {
    if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Too long since last message from root\r\n");
    if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Setting RootID to my ID\r\n");
    SyncStruct.masterRootId = Configuration.id;  // I'm the new root
    SyncStruct.syncPeriods = 0;
  }
  packet.to = BROADCAST;
  packet.type = SYNC_PACKET;
  packet.header_length = 0;
  packet.header = NULL;
  packet.data_length = sizeof (t);
  packet.data = (UINT8*) &t; 
  Net_send(&packet, NULL); 
  SyncStruct.syncPeriods++; 
  Timers_add(minuteDuration, sendSync, FRAME_LENGTH - 1);
  if(isRoot) {
    arrivalTimeSecs = currentTime.secs;
    arrivalTimeMillis = currentTime.millis;
    wakeUpTime = arrivalTimeSecs + (minuteDuration/1000) * FRAME_LENGTH;
    wakeUpTime -= secsAwakeBeforeSync;
    endSync();
  }
}

bool Synchronisation_init() {
  callbackSyncDone = 0;
  isRoot = 0;
  isInit = 1;
  if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Starts Synchonisation\r\n");
  SyncStruct.seqNum = 0;
  SyncStruct.masterRootId = Configuration.id;
  if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: New RootID = %i\r\n", Configuration.id);
  SyncStruct.syncPeriods = 0;
  Timers_add(0, sendSync, INIT_LENGTH);
  return True;
}

// handles a received sync message
bool Synchronisation_doSync() {
  if(isInit != 1) return True;
  if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Synch Packet received from %i\r\n", rxPacket.from);
  time_t* t1 = (time_t*) rxPacket.data;
  time_t* t2 = (time_t*) &rxTime;
  UINT8 sign;
  UINT32 diffs;
  UINT16 diffms;          
  // decide, if message is valid  
  // meaning from same root as before but newer
  // or from root with lower id as the old one
  UINT8  sn = rxPacket.num; //seqNum
  UINT16 mrid = rxPacket.from;  //masterRootId
  if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Sequence number received: %i\r\n", sn);
  if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: From RootID: %i\r\n", mrid);
  if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: My RootID: %i\r\n", SyncStruct.masterRootId);
  // message from invalid root
  if (SyncStruct.masterRootId < mrid) {
    if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Received root is higher than mine\r\n");
    SyncStruct.syncPeriods++;
    return True;
  // message from known root 
  } else if (SyncStruct.masterRootId == mrid) {
    if (SyncStruct.seqNum < sn) { // newer seqNum recieved
      //take newer message from same root as before
      SyncStruct.seqNum = sn;
      SyncStruct.syncPeriods = 0;
      if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Recieved newer message from known root\r\n");
    } else {  // older seqNum recieved
      if (mrid == Configuration.id) {  // but I'm the root
        isRoot = 1;
        SyncStruct.syncPeriods = 0;
      } else {
        SyncStruct.syncPeriods++;
      }
      if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Newer or same message already recieved from same root\r\n");
      return True;
    }
  } else {  // found a new master root
    if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Found new master root\r\n");
    if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: My ID is %i, new ID is %i\r\n", Configuration.id, mrid);
    SyncStruct.masterRootId = mrid;
    SyncStruct.seqNum = sn;
    SyncStruct.syncPeriods = 0;
  }
  // exact time the message arrived
  // Calculate difference
  if(t1->secs < t2->secs || (t1->secs == t2->secs && t1->millis < t2->millis)) { // t1 < t2
    sign = 1; // -
    diffs = t2->secs - t1->secs;
    // 6,2 8,4 -2,2      6,2 8,2 -2,0        6,2 8,1 -1,9
    if(t1->millis <= t2->millis) { diffms = t2->millis - t1->millis; }
    else { diffms = 1024 - (t1->millis - t2->millis); diffs--; }
  }
  else { // t1 > t2
    sign = 0;
    diffs = t1->secs - t2->secs;
    // 8,4 6,2 +2,2     8,2 6,2 +2,0    6,2 6,2 +0,0            8,1 6,2 +1,9
    if(t2->millis <= t1->millis) { diffms = t1->millis - t2->millis; }
    else { diffms = 1024 - (t2->millis - t1->millis); diffs--; }
  }
  if(Comm_logLevel & DEBUG_SYNC) Comm_print("SYNC: Diff seconds and diff milliseconds: %l,%i\r\n", diffs, diffms);
  // Adjust currentTime
  if(sign==0) {
    dint(); _NOP();
    currentTime.secs += diffs;
    currentTime.millis += diffms;
    if(currentTime.millis>1024) { currentTime.millis -= 1024; currentTime.secs++; }
      eint();
  } else {
    dint(); _NOP();
    currentTime.secs -= diffs;
    currentTime.millis -= diffms;
    if(currentTime.millis>1024) { currentTime.millis += 1024; currentTime.secs--; }
    eint();
  }
  arrivalTimeSecs = currentTime.secs;
  arrivalTimeMillis = currentTime.millis;
  // if we are kinda synchronized we can go sleep an wake up every x seconds
  if (diffs == 0) { // we have a sync error less than one second -> ok for our purposes
      // we received a sync message, next one will arrive in wakeUpTime seconds
      wakeUpTime = arrivalTimeSecs + (minuteDuration/1000) * FRAME_LENGTH;
      // dont miss sync msg, so wake up few seconds earlier
      wakeUpTime -= secsAwakeBeforeSync;
      // but before spread the new synchronization (for nodes outside of the root's area)
      // now go sleep
      endSync();
  }
  return True;
}

bool Synchronisation_registerSyncDone(fp_t fp) {
  callbackSyncDone = fp;
  return True;
}

bool Synchronisation_isRoot() {
  if(isRoot == 1) return True;
  else return False;
}
