/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/
 
/** \mainpage
 *
 * <p>
 * <h1> 1. Overview </h1>
 * <p> 
 * As the name already implies, sensor networks need sensors, typically lots of
 * sensors, to collect data from their environment. All sensor nodes may form an
 * ad-hoc network with some nodes acting as data sources, some as relays, and
 * some as data sinks collecting data. Nodes may act in all three roles at the
 * same time. 
 * <p>
 * The ESBs developed by the CST Group forms a sound basis for research in
 * sensor networks, the development of university level courses in CS/EE based
 * on real hardware, and prototyping of sensor net applications. 
 * <p>
 * The ESBs are part of the scatterweb approach comprising a whole family of 
 * embedded web servers, sensors and actors. Typical communication scenarios of 
 * sensor networks based on the ESBs are:
 * <p>
 * - ESBs communicate via the serial port with a standard PC for application
 *   development.
 * - ESBs communicate with mobile phones via the serial port to connect to
 *   wide-area mobile phone networks. This enables remote configuration of ESBs
 *   via short messages (SMS) as well as reception of sensor data on arbitrary
 *   mobile phones world-wide.
 * - ESBs communicate via their radio interface with other ESBs or an embedded
 *   web server (EWS) to form a truly embedded, highly flexible sensor network
 *   solution.
 * <p>
 * Main components of the ESB are the sensors, a micro controller for processing
 * of all data, and a transceiver handling communication with other nodes. 
 * <p>
 * Software for the following features is already available for the ESB 430:
 * <p>
 * - Configuration and analysis of all sensor data: vibration, noise, movement,
 *   temperature, infrared, (luminosity).
 * - Control of peripherals: timer, transceiver, serial port, EEPROM,
 *   IR-send/receive, LEDs, beeper, switches.
 * - Sending and receiving of data packets via the radio interface, transmission
 *   of sensor data, communication with other systems using the same radio
 *   interface (ESBs, ESWs).
 * - Sending and receiving of short messages (SMS) via connected mobile phone
 *   (serial port). Bluetooth connection available upon request (standard for
 *   the EWS). Just send the ESB an SMS triggering temperature monitoring every 
 *   Tuesday morning from 7:00 a.m. until 11:43 a.m., if and only if there is
 *   movement in the room but no light - it's possible with a single SMS!
 * - Sending (e.g. to control home entertainment devices) and receiving (e.g.
 *   from conventional remote controls) of RC5 packets via infrared.
 * - Periodic polling of sensor data and (depending on settings) automatic
 *   transmission via the radio interface, the serial port (terminal or mobile
 *   phone).
 * - Simple and easy configuration and control of the ESB via terminal commands
 *   over the serial interface or radio.
 * <p>
 * <h1> 2. Software Documentation </h1>
 * <p> 
 * The software of the sensor nodes was first seperated into two parts (firmware
 * & application) to provide in-field programming via radio. This has recently
 * changed because that reprogramming strategy created several problems.
 * Now the whole thing (except the reprogramming code, which should not change
 * anymore, see progamming.c) is reprogrammable. Anyway the seperation between
 * firmware and application is kept because it prooved good as hardware
 * abstraction layer and to provide some os-like concepts.
 * <p>
 * The Firmware provides abstract functions to interface and use the hardware,
 * contains the main execution loop + interrupts and provides some os-like
 * concepts. The firmware supports all ScatterWeb/MSP430 devices: 
 * ESB (and EYE), eGate, ECR.
 * The application (userapp, program) is built on top of the firmware.
 * <p>
 * Directory tree
 * - Applications: Contains application directories with PSPad project files.
 * - System:       Contains firmware sources and binaries.
 * <p>
 * <h1> 3. Major Changes </h1>
 * <p>
 * - Firmware and userapp are not that strictly separated anymore. This made
 *   reprogramming easier and all restrictions from the old firmware <-> userapp
 *   seperation are gone. Reprogramming code is now in programming.h, flasher.h
 *   contains the radio transfer of code images.
 * <p>
 * - For simplicity all 3 firmwares from ESB, EYE, eGate and ECR have been
 *   integrated into one with the use of #ifdef at some hardware dependent
 *   places. When compiling the command line one must always define one of: ESB,
 *   eGate, ECR, EYE. This is done in the makefiles with the variable DEVICE.
 *   The makefiles are now taking every source file they see in the directories
 *   and don't need to be changed for new/removed files anymore (Look into the
 *   makefiles and at new directory structure).
 * <p>
 * - Radio software has changed. Sending is now also done interrupt driven and
 *   comes with a ring buffer to store several packets to send. Additional new
 *   features are crc16 checksums and automatic acknowledgements (except
 *   broadcasts) with feedback to the sender of every packet.
 * <p>
 * - The terminal software has been moved to the firmware as many commands
 *   between devices were the same. Command creation has changed and most
 *   commands are executable from a remote device (with reply). Commands can now
 *   be placed where they are needed, Macros and a special section in flash 
 *   build a table to automatically search all commands (See
 *   ScatterWeb.Messaging.h).
 * <p>
 * - Light sensor software has been removed due to complex handling. Timer slots
 *   are also needed for constant dco checking&adjusting (ScatterWeb.Timers.h)
 *   and a software RTC (ScatterWeb.Time.h) as the RTC is mostly abandoned
 *   because of its big current usage, it's only used as an well calibrated
 *   temperature sensor (ScatterWeb.System.h).
 * <p>
 * - ScatterWeb.Time.h represents the new software RTC (see currentTime). 
 *   ScatterWeb.System.h has been completly cleared, only the function for
 *   reading the temperature and ensuring RTC 32 khz output at initialization
 *   are left. fiveMsTimer is replaced through currentTime or use of timers
 *   (ScatterWeb.Timers.h), its still left for internal timer use.
 * <p>
 * - ScatterWeb.Timers.h module has been added, no need to poll for an elapsed
 *   time anymore. ScatterWeb.System.h includes DCO checking, so DCO is
 *   recalibrated the whole time, not at startup only!
 * <p>
 * - Other modules have been more seperated (ScatterWeb.Configuration.h (was in
 *   ScatterWeb.IO.h), programming.h (was in flasher.h)).
 * <p>
 * - The watchdog has been rewritten. It's now correct and possible to see where
 *   a reset came from. Additionaly a feature has beend added to watch stack
 *   growth and detect a stack overflow. For bad conditions like that a panic
 *   function has been added. The firmware can disable the application if a
 *   panic occured.
 * <p>
 * - An abstract printf like function has been implemented for string
 *   outputting: String_abstractWrite. This is used by the functions
 *   String_write (to write a string to memory), Comm_print (to write normal
 *   output to the serial port) and Comm_log (to write debug output to the
 *   serial port). No other methods should be used for writing to the serial
 *   port, sendRS232 is not public anymore. This is because serial port output
 *   is now controlled via Comm_logLevel. It allows several log levels, some
 *   debug bits for special debugging and total disabling of all output (see
 *   ScatterWeb.Comm.h).
 * <p>
 * - A cooperative scheduler exists to create simple task functions:
 *   ScatterWeb.Threading.h.
 * <p>
 * - The main superloop goes to sleep when everything is processed, all things
 *   that need to be handled are interrupt driven (see main). Every interrupt
 *   which needs processing from the superloop wakes the processor up with
 *   WAKEUPLPM1, which also sets a flag in runModule. When all flags in
 *   runModule are reset (after processing from superloop tasks) the loop goes
 *   to LPM1 again. If a function which needs calls as often as possible is
 *   added to the superloop you have to comment out the lines about sleeping! 
 * <p>
 * - Typedefs for UINT8, UINT16, UINT32 have been introduced because it is
 *   better style and much shorter.
 * <p>
 * <h1> 4. Important things to know </h1>
 * <p> 
 * - All routines should return fast. If a routine occupies the processor for a
 *   long time (for example outputting a long text which works with the blocking
 *   sendRS232Char() function) it should call System_startWatchdog() repeatedly
 *   (normally done in main()) to avoid a watchdog reset. If a watchdog reset
 *   occurs the firmware will disable the programm if PANIC_DISABLES_USERAPP is
 *   set.
 * <p>
 * - You don't need to change the makefiles anymore. Makefiles in firmware and
 *   application automatically compile every .c file in the src directories and
 *   look into every .c and .h file for changes.
 * <p>
 * - No intialization of globals is done. You can write "UINT16 var = 49;" but
 *   var will never be initialized, do initialization in code! (Initialization
 *   may work again if the standard startup routine is used.)
 * <p>
 * - Don't use dynamic memomry allocation (malloc/free)! Dynamic memory
 *   allocation bears the following problems on MCUS:
 *   - Much harder to debug.
 *   - Unknown behaviour at compile time.
 *   - Memory fragmentation and no way (missing MMU) to defragment it.
 *   - No memory protection between stack and heap which grow towards each
 *     other.
 *   - Additionally current msp430-gcc malloc implementation seems to be buggy.
 * <p>
 * Sometimes an embedded engineer said:
 * "Think small on MCUs. If you are planning to use dynamic memory allocation,
 * you are thinking much too BIG."
 * <p>
 * - If weird things happen look into linker map "out.map"
 * <p>
 * - Sometimes compilation fails or hangs with existing objects files from
 *   another machine. Do a "make clean" or simply delete all object (*.o) files
 *   in firmware and userapp directories in this situation!
 */  

/** \file ScatterWeb.System.h
 ** \ingroup System
 ** \brief The ScatterWeb.System contains fundamental functions that define
 ** commonly-used value and reference data types, events and event handlers,
 ** attributes, and processing exceptions.
 ** <p>
 ** ADC: Starting, enabling, disabling and ISR for ADU.
 ** The main function of the ADU is to sample the unfiltered radio receive
 ** output and provide a "carrier detect" (see interrupt routine ADC12ISR and
 ** rxReceiveLimit in Configuration ).   
 ** <p>
 ** Look for the defines in ScatterWeb.System.h to access other adc readings.
 ** <p> 
 ** CLOCK: Methods for interfacing DS1629 (RTC with temperature) and LM71
 ** (temperature sensor).
 ** <p>
 ** For both devices the function getTemperature is implemented. It returns the
 ** temperature in fixed point [9.7] 2s complement format!  
 ** <p>   
 ** DS1629
 ** initClock should be called at startup to check for the correct setting in the
 ** status register of the RTC.
 ** <p>
 ** Most functions of the clock have been removed due to high current usage
 ** while reading the clock. Furthermore a software RTC has been implemented
 ** with a much finer resolution (see ScatterWeb.Time.c). The RTC is now only
 ** used as an well calibrated temperature sensor and only these functions
 ** remain here (for complete clock interfacing code look into attic directory).    
 ** <p>
 ** cl_start(), cl_stop(), cl_writeOnBus() and cl_readFromBus are the basic
 ** functions for interfacing the RTC (I2C like bus).
 ** <p>
 ** Ports (ESB/EYE): clockSDA P50, clockSCL P51, clockAlarm P15
 ** <p>
 ** Note: The sticks have no RTC and the ECR has the LM71 temperature sensor
 **       which is also realized in this file.  
 ** <p>
 ** CRC: This file offers crc calculation.
 ** <p>
 ** DCO: Initalizing & Controlling the DCO (Digitally Controlled Oscillator)
 ** <p> 
 ** If FF_DCOCHECKER in config_t.firmwareFlags is set Timers_init will add a
 ** timer every 100 ms for checkDCO and enable Timer_B1 interrupt. 
 ** <p>
 ** The interrupt continually measures the ACLK with the DCO (like in 
 ** System_setDCO) and writes the result into dcoDiff. checkDCO takes samples
 ** every 100 ms of dcoDiff and adjusts the DCO if needed.
 ** <p>
 ** As the DCO is temperature-sensitive and both serial and radio communication
 ** depend on the DCO calibrating the DCO continously is recommended. Serial
 ** communication stops working at a temperature difference of ~10C.
 ** <p>
 ** A more low power option might be to not use checkDCO and just call
 ** System_setDCO in long intervals (a minute or even longer).     
 ** <p> 
 ** WATCHDOG: Hardware&Software watchdog, reset reason, stack check and panics.
 ** <p>
 ** This module contains two macros to control the hardware watchdog.
 ** Furthermore it supports determining the reset source, a simple stack checker
 ** to avoid stack overflows and a panic function.
 ** <p>
 ** Hardware & Software Watchdog
 ** The macro System_startWatchdog() allows starting or resetting both
 ** watchdogs, System_stopWatchdog() stops it. These macros are mostly for
 ** firmware internal use. In the perfect case noone should call
 ** System_stopWatchdog() and the only place to call System_startWatchdog()
 ** repeatedly would be the main loop. However some software parts need a long
 ** time or must run exclusively (then also disabling the interrupts together
 ** with System_stopWatchdog() )  
 ** <p>
 ** Timer_A1 Interrupt was used to implement a software watchdog with a timer
 ** interrupt running together with the built in hardware watchdog, because an
 ** interrupt offers the possibility to receover the address where the watchdog
 ** timeout occured. This address gets written in a panic message, when Timer_A1
 ** Interrupt fires (what means watchdog timeout). Putting the hardware watchdog
 ** into interrupt mode (what is possible with the MSP430) is not a good idea,
 ** because then it can't recover the MCU in a case, where it is not running at
 ** all (for example MSP430s known startup problems with a slow or noisy voltage
 ** rise which lead to such a situation).             
 ** <p>
 ** Reset source
 ** <p>
 ** The function wd_init() runs at the very beginning of main just after
 ** initPort(). It checks some registers to see where the reset came from and
 ** then sets wd_startup acoordingly. 
 ** <p>
 ** Technical details of wd_init():      
 ** The MSP internally has two reset signals (slau049.pdf, Chapter 3):
 ** - POR (power-on reset): This is a device reset. Set by Power-Up or a low RST
 **                         line when configured in reset mode (standard).   
 ** - PUC (power-up clear); A PUC follows every POR. But it can occur also
 **                         without because of watdchdog timer expiry or flash
 **                         control register key violation.
 ** <p>
 ** The SFRs are initialized on resets and some bits get only set or reset, when
 ** a POR occured, not on a PUC only (these in braces, see slau049.pdf, page
 ** 3-4).
 ** <p>
 ** This feature is used by wdt_init with the register TACTL to detect a POR
 ** since it will only get reset (complete zero) if a POR occured.    
 ** <p>
 ** If a PUC only occured it was a warm start. The function looks further wether
 ** it was the WDT, Flash key violation or an unknown source. If it was a Flash
 ** key violation it is checked with var wd_state wether it was wanted by
 ** software (System_reset() with wd_state == WDS_SWRESET or System_panic() with
 ** wd_state == WDS_PANIC) or a real key violation.
 ** <p>
 ** Stack checking
 ** wd_stackCheck is a small function which once called by main runs once every
 ** second with a timer. It checks 8 bytes on top of the data segment (behind 
 ** __noinit_end provided by the linker script) for corruption. The reset_vector
 ** in main() fills the memory from __noinit_end - 0x09FF with 0xABCD, so the
 ** procedure simply checks the four ints from __noinit_end for the value
 ** 0xABCD. If they are not there, the stack must have gone that low and
 ** possibly lower. The procedure will panic then.   
 ** <p>
 ** Panic
 ** System_panic is a function to call when some bad condition was detected (for
 ** example low stack, full timers). System_panic writes a small info msg into
 ** second info memory segment (0x1080-0x10FF) and then resets the controller. 
 ** Multiple panic messages can be written until the segment is full. 
 ** <p>
 ** One could go further and implement a complete memory dump into flash, what
 ** would give the possibility to examine the complete system state later for
 ** debugging.
 ** <p>  
 ** LPM: Power saving
 ** The sensorboard features low power modes to reduce power consumption. This
 ** increases battery-powered lifetime up to years and makes it even possible to
 ** run the sensorboard solely with a goldcap/solar cell combo.
 ** <p>
 ** Following units can be switched off to reduce power consumption:
 ** <ul>
 ** <li>Unit - Power consumption - Corresponding function - Comments
 ** <li>ScatterWeb.Net - 4,7 - 5,2 ma - Net_off(), Net_on()
 ** <li>ScatterWeb.Data - 2,9 ma - System_powerOff(), System_powerOn()
 ** </ul>
 ** <p>
 ** The MSP430 itself goes to sleep when nothing more has to be done (see super
 ** loop in main). Power consumption with everything disabled (Radio, Sensor)
 ** goes below 1 mA. 
 ** <p> 
 ** This file contains special definitions & configuration options for the
 ** firmware and includes all other headers.
 ** <p>
 ** The firmware provides abstract functions to interface and use the hardware,
 ** contains the main execution loop + interrupts and provides some os-like
 ** concepts. The firmware supports all ScatterWeb/MSP430 devices: 
 ** ESB, EYE, eGate, ECR
 ** <p>
 ** The first reprogramming concept which produced the firmware/application
 ** seperation was abandoned for two main reasons:
 ** <p>
 ** - It relied on a stable almost never changing firmware. At the moment this
 **   is absolutely wrong and its questionable if it will ever be.
 ** - It relied on the compiler/linker to produce always the same assembler
 **   output for the firmware when it gets always the same sources/object files.
 **   This was only partially right (it depended on too many factors: compiler
 **   version, compiler flags, even the machine the compiler is running on, ...)
 **   and could be wrong under other or future conditions.
 ** <p>
 ** Perhaps one needs to write his own compiler/linker (or at least modify an
 ** existing one) to really seperate these two software parts on a MCU cleanly.
 **
 ** @author        Marco Hutta
 ** @author        Bjoern Lichtblau
 ** @author        Thomas Pietsch 
 ** @version       2.3
 ** @since         2.0
 **/

#ifndef ScatterWeb_System
#define ScatterWeb_System

/**
 * Define WAKEUPLPM1
 *
 * @since         2.0
 */
#define WAKEUPLPM1(irqname, bit)  do {     \
    __asm__ __volatile__ ("bic %0, .L__FrameOffset_" #irqname "(r1)": : "i" ((uint16_t)LPM1_bits)); \
    runModule |= bit; \
    }while(0)

/**
 * Check if any device was specified, should be done in the makefiles with var
 * $DEVICE.
 *
 * @since         2.0
 */
#if !(defined(ESB) || defined(eGate) || defined(ECR))
#error "No device specified: ECR, ESB, eGate"
#endif

/**
 * Buffer size: Maximum size for normal terminal replies.
 *
 * @since         2.0
 */
#define TERM_BUFSIZE        (82)

/**
 * Buffer size: Maximum size of a line received by uart, rest will be cut.
 *
 * @since         2.0
 */
#define SERIAL_BUFSIZE      (82)

/**
 * Buffer size: Maximum size of packets that can be received, header included is
 * 8 bytes.
 *
 * @since         2.0
 */
#define RADIO_RXBUF_SIZE    (270)

/**
 * Buffer size: Size of the TX ring buffer.
 *
 * @since         2.0
 */
#define RADIO_TXBUF_SIZE    (400)

/**
 * Buffer size: Maximum number of timers, each timer requires 6 bytes of RAM
 * (==sizeof(timerEntry)).
 *
 * @since         2.0
 */
#define MAX_TIMERS          (20)

/**
 * Buffer size: Maximum number of tasks, each entry is 10 bytes.
 *
 * @since         2.0
 */
#define MAX_TASKS           (10)

/**
 * Define special section definition for linking
 *
 * @since         2.0
 */
#define INFOMEM            __attribute__((section(".infomem")))

/**
 * Define special section definition for linking
 *
 * @since         2.0
 */
#define INFOMEMNOBITS      __attribute__((section(".infomemnobits")))

/**
 * Standard bit manipulation macro.
 *
 * @since         2.0
 */
#define   SET(x, y)     ((x) |= (y))

/**
 * Standard bit manipulation macro.
 *
 * @since         2.0
 */
#define   CLEAR(x, y)   ((x) &= (~y))

/**
 * Module flag for low power management: Set if Threading_eventHandler should be
 * called.
 *
 * @since         2.0
 */
#define MF_SCOS       (0x0001)

/**
 * Module flag for low power management: Set if Net_rxHandler should be called.
 *
 * @since         2.0
 */
#define MF_RADIO_RX   (0x0002)

/**
 * Module flag for low power management: Set if Net_txHandler should be called.
 *
 * @since         2.0
 */
#define MF_RADIO_TX   (0x0004)

/**
 * Module flag for low power management: Set if Timers_eventHandler should be
 * called.
 *
 * @since         2.0
 */
#define MF_TIMER      (0x0008)

/**
 * Module flag for low power management: Set if a serial line is complete.
 *
 * @since         2.0
 */
#define MF_SERIAL_RX  (0x0010)

/**
 * Module flag for low power management
 *
 * @since         2.0
 */
#define MF_SENSORS    (0x0020)

/**
 * Module flag for low power management: Set if a RC5 code is received via IR
 * Receiver.
 *
 * @since         2.0
 */
#define MF_RC5        (0x0080)

/**
 * Define: Power on or reset button.
 *
 * @since         2.0
 */
#define STARTUP_POR       (0)

/**
 * Define: call via System_reset().
 *
 * @since         2.0
 */
#define STARTUP_SWRESET   (1)

/**
 * Define: reprogramming
 *
 * @since         2.0
 */
#define STARTUP_REPROGRAM (2)

/**
 * Define: call via System_panic(). All states <=STARTUP_PANIC are considered
 * OK!
 *
 * @since         2.0
 */
#define STARTUP_PANIC     (3)

/**
 * Define: Reason for reset was watchdog expiry.
 *
 * @since         2.0
 */
#define STARTUP_WATCHDOG  (4)

/**
 * Define: Flash control register key violation.
 *
 * @since         2.0
 */
#define STARTUP_KEYV      (5)

/**
 * Define: Unknown type of reset.
 *
 * @since         2.0
 */
#define STARTUP_UNKNOWN   (6)

/**
 * Define: System at startup.
 *
 * @since         2.0
 */
#define WDS_RUNNING       (0xAB01)

/**
 * Define: Software sets this shortly before writing bad value into FCTL3 to
 * recognize a wanted reset.
 *
 * @since         2.0
 */
#define WDS_SWRESET       (0xAB02)

/**
 * Define
 *
 * @since         2.0
 */
#define WDS_REPROGRAM     (0xAB03)

/**
 * Define
 *
 * @since         2.0
 */
#define WDS_PANIC         (0xAB04)

/**
 * Define: Initial FCS value
 *
 * @since         2.0
 */
#define INIT_FCS  0xffff

/**
 * Callback type: Gets called everytime a whole line is received via serial port.
 *
 * @see           #System_registerCallback()
 * @since         2.0
 */
#define C_SERIAL              1

/**
 * Callback type: Gets called everytime an unknown packet (not handled by the
 * firmware) is received.
 *
 * @see           #System_registerCallback()
 * @since         2.0
 */
#define C_RADIO               2

/**
 * Callback type: Gets called for sensor values.
 *
 * @see           #System_registerCallback()
 * @since         2.0
 */
#define C_SENSOR              3

/**
 * The callbacks[] Array will get this size.
 *
 * @see           #callbacks[] 
 * @since         2.0
 */
#define NUMBER_CALLBACKS	    4 

/**
 * Define
 *
 * @since         2.0
 */
#define MICIFG    (0x01)

/**
 * Define
 *
 * @since         2.0
 */
#define MICIE     (0x02)

/**
 * Define
 *
 * @since         2.0
 */
#define MICVOLTAGE        (ADC12MEM2)

/**
 * Define
 *
 * @since         2.0
 */
#define EXTERNVOLTAGE     (ADC12MEM3)

/**
 * Define
 *
 * @since         2.0
 */
#define RXPVOLTAGE        (ADC12MEM5)

/**
 * Define
 *
 * @since         2.0
 */
#define BATTERYVOLTAGE    (ADC12MEM4)

/**
 * Enum
 * <ul>
 * <li>True
 * <li>False
 * </ul> 
 *
 * @see           #bool 
 * @since         2.0
 */
enum {
  True = 1,
  False = 0,
};

/**
 * Bool
 * 
 * @since         2.0
 */
typedef char bool;

/**
 *  Temperature type (built on a signed int). It's a signed (twos complement) 
 *  fixed point value with 8 bits before comma and 7 bits after. So Bit 15 is
 *  sign, Bit14-7 is before comma and Bit 6-0 after comma. 
 * 
 * @since         2.0
 */
typedef union { unsigned int u; signed int s; } temp_t; 

/**
 * Elementary data type
 * 
 * @since         2.0
 */
typedef unsigned char     UINT8;

/**
 * Elementary data type
 * 
 * @since         2.0
 */
typedef unsigned int      UINT16;

/**
 * Elementary data type
 * 
 * @since         2.0
 */
typedef unsigned long     UINT32;

/**
 * Several predefined function pointer type: For functions like: void
 * function(void).
 * 
 * @since         2.0
 */
typedef void(*fp_t)();

/**
 * Several predefined function pointer type: For functions like: form: void
 * function(UINT8*)
 * 
 * @since         2.0
 */
typedef void(*fp_string_t)(UINT8*);

/**
 * Several predefined function pointer type: For functions like: void
 * function(UINT8)
 * 
 * @since         2.0
 */
typedef void(*fp_char_t)(UINT8);

/**
 * Several predefined function pointer type: For functions like: void
 * function(UINT16)
 * 
 * @since         2.0
 */
typedef void(*fp_int_t)(UINT16);

/**
 * Several predefined function pointer type: For functions like: void
 * function(void*).
 * 
 * @since         2.0
 */
typedef void(*fp_vp)(void*);

/**
 * Table where callbacks are hold
 *
 * @since         2.0
 */
extern fp_vp callbacks[];

/**
 * UINT16
 *
 * @since         2.0
 */
extern volatile UINT16 runModule;

/**
 * Last updated RXPower AD value.
 *
 * @since         2.0
 */
extern volatile UINT16 rxpTemp;

/**
 * RXPower AD value for last received packet.
 *
 * @since         2.0
 */
extern volatile UINT16 rxpValue;

/**
 * 0x01 is like IFG, 0x02 like IE and will set sensorFlags.MIC
 *
 * @since         2.0
 */
extern volatile UINT8 MIC;

/**
 * Registers callback
 *
 * @param type    type of callback
 * @param ()      function to call 
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool System_registerCallback(UINT8 type, void(*fp_t)()); 

/**
 * Returns temperature
 *
 * @return        current temperature.
 * @since         2.0
 */
temp_t System_getTemperature();

/**
 * Sets DCO
 *
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool System_setDCO();

/**
 * Starts watchdog
 *
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool System_startWatchdog();

/**
 * Stops watchdog
 *
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool System_stopWatchdog();

/**
 * Resets System
 *
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool System_reset();         

/**
 * Checks the reset
 *
 * @return        pointer of reset message.
 * @since         2.0
 */
const UINT8* System_resetReason();

/**
 * Sets panic flag
 *
 * @param fmt     panic to set 
 * @param ...     parameters to set
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool System_panic(const char *fmt, ...);

/**
 * Checks panic flag
 *
 * @param n       panicReason[n]
 * @return        pointer of the panic messages.
 * @since         2.0
 */
UINT8* System_panicReason(UINT16 n);

/**
 * Clears the panic flags
 *
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool System_panicClear();

/**
 * Waits - A simple looping waiting function. Rough estimate at 2,4576 MHz: w=1
 * => 18us, w=10 => 62us, w=100 => 500us  w=400 => 4ms.
 * This method has two big disadvantages: it blocks and it's not accurate,
 * timers are better.
 *
 * @param w       ticks to wait 
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool System_wait(UINT32 w);

/**
 * Waits which is more precise than wait
 * because it uses TimerA to produce its interval, also blocking.
 *
 * @param i       ticks to wait
 * @return        <code>True</code> always.
 * @since         2.0
 */ 
bool System_waitA(UINT16 i);

/**
 * Enables the sensor and serial power
 *
 * @return        <code>True</code> always.
 * @since         2.2
 */  
bool System_powerOn();

/**
 * Disables the sensor and serial power
 *
 * @return        <code>True</code> always.
 * @since         2.2
 */  
bool System_powerOff();

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
UINT16 System_reprogram();

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool System_startConversion();

// Include all headers here so other files only need to include this file.
#include <msp430x14x.h>
#include <io.h>
#include <signal.h>
#include <stdlib.h>
#include <stdarg.h>

#if defined(ECR)
#include "../lib/ecr/programming.h"
#include "../lib/ecr/flasher.h"
#elif defined(ESB)
#include "../lib/esb/programming.h"
#include "../lib/esb/flasher.h"
#elif defined(eGate)
#include "../lib/egate/programming.h"
#include "../lib/egate/flasher.h"
#endif

#include "ScatterWeb.Configuration.h"
#include "ScatterWeb.Data.h"
#include "ScatterWeb.Timers.h"
#include "ScatterWeb.Time.h"
#include "ScatterWeb.IO.h"
#include "ScatterWeb.Comm.h"
#include "ScatterWeb.Net.h"
#include "ScatterWeb.Messaging.h"
#include "ScatterWeb.Threading.h"
#include "ScatterWeb.String.h"
#include "ScatterWeb.Synchronisation.h"

#endif
