/*
Copyright 2005, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2005
*/

/** \file ScatterWeb.Threading.h
 ** \ingroup System
 ** \brief The ScatterWeb.Threading provides functions that enable multithreaded
 ** programming.
 ** <p>
 ** ScatterOS
 ** This file implements a simple cooperative scheduling mechanism which allows
 ** to run task functions with os like blocking calls (which in fact do not
 ** block the processor but return control to the system and get later resumed).
 ** <p>
 ** The os primitives at the moment are:
 ** <ul>
 ** <li>Threading_doEvents - Simply returns control to the system. As this is a cooperative scheduler long tasks or tasks which make no other Threading_ calls must interrupt themself with Threading_doEvents.
 ** <li>Threading_sleep - Makes the task wait a specified amount of system ticks (may get longer if system is busy).
 ** <li>Threading_wait - Makes the task wait for a flag set.
 ** <li>Threading_recv / Threading_recvp - Makes the task wait for a packet of a given type (and a given sender if wanted). A timeout can be set and Threading_recvp works in promiscous mode.
 ** <li>Threading_send / Threading_esend - A synchronous send, task will continue after the packet was sent or lost (removed from radio rx ringbuffer). Threading_esend is for sending data from eeprom.
 ** </ul>
 ** <p>
 ** These are not many but suit the needs at the moment. 
 ** <p>
 ** todo: Threading_ calls for waiting for sensor readings and some methods for
 ** task synchronisation (binary semaphores, counting semaphores or even more?!)
 ** <p>
 ** Constraints
 ** <p>
 ** The goal was to write a task switching mechanism with minimal memory
 ** requirements. Each preemptive OS requires and most cooperative OS use one
 ** stack per task (Exceptions seen so far are PeerOS in the EYES Porject, which
 ** preempts only lower priority tasks and lets them reside at the stack while
 ** executing a higher priority task, and Salvo, which seems to implement a very
 ** similar approach).
 ** <p>
 ** To have a stack for each task increases memory requirements a lot, because
 ** the size of the stack for a task is unknown and must be generously set. The
 ** optimum would be the maximum stack depth of the task call tree. A compiler
 ** could easily calculate this if recursion and function pointers are not used,
 ** but i do not know if any compiler supports this feature. The second bad is
 ** that most times a task won't use its maximum stack depth.
 ** <p>
 ** In contrast to theses approaches this mechanism uses no memory (only some
 ** static allocated memory for task table and a little bit of stack space like
 ** in every other function is needed). This puts some limits onto the task
 ** functions to make this possible. 
 ** <p>
 ** Task functions can't and should never allocate stack space. In fact you can
 ** allocate stack space and the scheduler will handle it properly (that was
 ** needed for calls with long argument list which get pushed onto the stack),
 ** but after an Threading_ call the stack content will always be undefined. 
 ** <p>
 ** That means that all state must be held in vars declared globally.
 ** Additionally they should be declared volatile so the compiler won't do any
 ** optimization on them (hold them in registers, also register content is
 ** undefined after an Threading_ call). Secondly it means that blocking 
 ** Threading_ function are only allowed to call within your top-level task
 ** function, not in functions called from your top-level task function! The
 ** reason for that is again the stack, which is undefined after an Threading_
 ** call and so will thrash the return address for Threading_ calls anywhere
 ** outside from your top-level task function.
 ** <p>
 ** How it works
 ** A task is initialized with Threading_add, which gets a pointer to the task
 ** function. If a task can run (like a newly added task) sooner or later
 ** scos_call with the id of the task in the task table and a value is called. 
 ** scos_call saves the current stack pointer, sets currentTask and calls the 
 ** next-to-continue address of the task like a function with an int argument.
 ** <p>
 ** Then the task runs until it calls an Threading_ function. These functions
 ** save the return address of the task on the stack as the next-to-continue
 ** address in the task table, change the state of the task and call scos_exit.
 ** <p>
 ** scos_exit now just restores the stack pointer to the value scos_call saved
 ** and does a return, which now is like the return of scos_call and so the
 ** system is again in a clean state.
 ** <p> 
 ** Related
 ** From slau049.pdf / Page 5-2:
 ** 5.1.2 The System Stack Pointer (SP)
 ** The system stack pointer must always be aligned to even addresses because
 ** the stack is accessed with word data during an interrupt request service.
 ** The system SP is used by the CPU to store the return addresses of subroutine
 ** calls and interrupts. It uses a predecrement, postincrement scheme. The
 ** advantage of this scheme is that the item on the top of the stack is
 ** available. The SP can be used by the user software (PUSH and POP
 ** instructions), but the user should remember that the CPU also uses the SP.
 ** Figure 52 shows the system SP bits.
 ** <p>
 ** From slau049.pdf / Page 5-19, Call a subroutine:
 ** CALL dst          SP  2 > SP, PC+2 > stack, dst > PC
 ** <p>
 ** From slau049.pdf / Page 5-22, Return from subroutine:
 ** RET  MOV @SP+,PC
 ** <p>
 ** From the GCC manual:
 ** -fno-defer-pop 
 ** Always pop the arguments to each function call as soon as that function
 ** returns. For machines which must pop arguments after a function call, the
 ** compiler normally lets arguments accumulate on the stack for several
 ** function calls and pops them all at once. 
 ** <p>
 ** Perhaps a cleaner and more portable but also more memory consuming
 ** implementation would be possible with setjmp()/longjmp() macros, see 
 ** - http://publications.gbdirect.co.uk/c_book/chapter9/nonlocal_jumps.html
 ** - http://www.csl.mtu.edu/cs4411.ck/www/NOTES/non-local-goto/mtp.h
 ** <p>
 ** Helper for scos_call. 
 ** Had to be put into an extra functions because of following problem:
 ** The following instruction ("mov &savedSP, r1"::) is not really needed
 ** because if execution returns here, the stack poitner should be what we saved
 ** before. The important thing is to tell the compiler, that the usually caller
 ** preserved registers (r6-r11) may have changed! The problem is the indexing
 ** into the Tasks table. If following line is omitted the assembler output is:
 ** <p>
 ** scos_prolog(value, Tasks[currentTask].address, Tasks[currentTask].stackSize);
 ** 7154:	1e 42 50 06 	mov	&0x0650,r14	;0x0650
 ** 7158:	3b 40 54 06 	mov	#1620,	r11	;#0x0654
 ** ...
 ** 718a:	b0 12 42 71 	call	#28994		;#0x7142
 ** //__asm__ __volatile__("mov &savedSP, r1"::); // restore stack pointer
 ** Tasks[currentTask].address = 0x0000;
 ** 718e:	1e 42 50 06 	mov	&0x0650,r14	;0x0650
 ** ...
 ** 71a0:	0f 5b       	add	r11,	r15	;
 ** 71a2:	8f 43 00 00 	mov	#0,	0(r15)	;r3 As==00
 ** Tasks[currentTask].state = T_DESTROYED;
 ** ...
 ** 71b8:	0f 5b       	add	r11,	r15	;
 ** 71ba:	8f 43 04 00 	mov	#0,	4(r15)	;r3 As==00
 **    
 ** That is bad. At the beginning the base address of Tasks is moved into r11
 ** for reading out of the table for the scos_prolog call. r11 is later
 ** (after scos_prolog call) used to index into Tasks again, and the compiler
 ** can assume, that r11 did not change (according to normal calling
 ** conventions), but that may be wrong in this case.
 ** <p>
 ** When the following instruction is used, the compiler omits the "add r11,r15"
 ** and addresses the table directly:
 ** <p> 
 ** __asm__ __volatile__("mov &savedSP, r1"::); // restore stack pointer
 ** 718e:	11 42 4a 06 	mov	&0x064a,r1	;0x064a
 ** Tasks[currentTask].address = 0x0000;
 ** ...
 ** 71a4:	8f 43 54 06 	mov	#0,	1620(r15);r3 As==00
 ** Tasks[currentTask].state = T_DESTROYED;
 ** ...
 ** 71ba:	8f 43 58 06 	mov	#0,	1624(r15);r3 As==00
 ** <p>
 ** I dont know why the compiler does that thing not always this way (it saves
 ** on instruction). I tried to force it by telling all registers get clobbered
 ** through this instruction:
 ** <p>
 ** __asm__ __volatile__("mov &savedSP, r1":: : "r15", "r14", "r13", "r12",
 ** "r11", "r10", "r9", "r8", "r7", "r6"); 
 ** restore stack pointer
 ** <p>
 ** But this produced the same behaviour like the first, BUT instead of r11
 ** being used it is now r5 which is even more worse. r5 is the frame pointer
 ** and strangely used by gcc. The compiler should never assume r5 is preserved
 ** over function calls, very strange why it does that here (even when r15 is
 ** appended to the list of clobbered registers).
 ** <p>
 ** Anyway it works with the second. But that may change with -O flags or
 ** anything else, so best will be to put that into an extra function. <=== DONE
 ** <p>
 ** Make sure this function does not get inlined into scos_call again with
 ** -fno-inline for this file ?! 
 **
 ** @author        Marco Hutta
 ** @author        Bjoern Lichtblau
 ** @author        Thomas Pietsch 
 ** @version       2.3
 ** @since         2.0
 **/  
 
#ifndef ScatterWeb_Threading
#define ScatterWeb_Threading

/**
 * current task
 *
 * @since         2.0
 */
extern volatile UINT16 Threading_currentThread;

/**
 * Initializes the threading
 *
 * @return        <code>True</code> always.
 * @since         2.0
 */
bool Threading_init();

/**
 * Adds thread
 *
 * @param address   address of task
 * @param number    number to write
 * @param length    length to write
 * @return          <code>True</code> if successful;
 *                  <code>False</code> otherwise. 
 * @since           2.0
 */
bool Threading_add(void* address);

/**
 * Does events
 *
 * @return          <code>True</code> always.
 * @since           2.0
 */
bool Threading_doEvents();

/**
 * Sleeps for ticks
 *
 * @param ticks     ticks to wait
 * @return          <code>True</code> always.
 * @since           2.0
 */
bool Threading_sleep(UINT16 ticks);

/**
 * Waits until address == bitmask
 *
 * @param adr       address
 * @param bitmask   bitmask
 * @return          <code>True</code> always.
 * @since           2.0
 */
bool Threading_wait(UINT8* adr, UINT8 bitmask);

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Threading_eventHandler(); 

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Threading_packetHandler();

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Threading_sentPacketHandler(UINT16 p, UINT16 success);

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Threading_recv(UINT16 sender, UINT16 type, UINT16 timeout);

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Threading_recvp(UINT16 sender, UINT16 type, UINT16 timeout);

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Threading_send(UINT16 to, UINT16 type, UINT8* h, UINT16 hl, UINT8* d, UINT16 dl);

/* 
 * This member supports the ScatterWeb infrastructure
 * and is not intended to be used directly from your code.
 */
bool Threading_esend(UINT16 to, UINT16 type, UINT8* h, UINT16 hl, UINT16 d, UINT16 dl);

#endif
